"""Contain the encrypted socket class."""
import logging
import socket
import threading
import time
import ssl

from pathlib import Path

from typing import Any
from typing import Callable
from typing import Optional
from typing import Union

from .protocol import StatusID
from .protocol import Connection

from ..utils import observer


class TlsSocket(Connection):
    """Handle the encrypted socket connection."""

    def __init__(
        self,
        address: str,
        port: int,
        ca: Path,  # ca.crt
        crt: Path,  # client.crt
        key: Path,  # client.key
    ):
        """."""
        self.log = logging.getLogger(__name__)
        self.log.addHandler(logging.NullHandler())

        self.observer_name = "CONNECTION"
        self.observer = observer
        self.observer.post(StatusID.DISCONNETCED, None)

        self.send_ready = threading.Lock()

        self.address = address
        self.port = port
        self.socket_timeout = 30_000
        self.RECEIVE_SIZE = 2048
        self.killed = threading.Event()

        self.log.debug(f"Address: {self.address}")
        self.log.debug(f"Port: {self.port}")

        self.ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        self.ssl_context.check_hostname = True
        self.ssl_context.verify_flags = ssl.OP_NO_TLSv1_1
        self.ssl_context.verify_mode = ssl.CERT_REQUIRED
        # if logging.root.level <= logging.DEBUG:  # NOTE: Only works after 3.8
        #     self.ssl_context.keylog_filename = "keylog_file.log"

        self.ssl_context.load_cert_chain(certfile=crt, keyfile=key)
        self.ssl_context.load_verify_locations(cafile=ca)

    def _socket_setup(self) -> None:
        """
        Create socket to communicate with server.

        Creates a socket instance and sets the options for communication.
        Passes the socket to the ssl_wrap method

        Note:
        After 5 idle minutes, start sending keepalives every 1 minutes.
        Drop connection after 2 failed keepalives
        """
        self.raw_socket: Optional[socket.socket] = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.raw_socket.setsockopt(
            socket.SOL_SOCKET,
            socket.SO_KEEPALIVE,
            1
        )
        self.raw_socket.settimeout(2)
        if (
            hasattr(socket, "TCP_KEEPIDLE")
            and hasattr(socket, "TCP_KEEPINTVL")
            and hasattr(socket, "TCP_KEEPCNT")
        ):
            self.log.debug(
                "Setting TCP_KEEPIDLE, TCP_KEEPINTVL & TCP_KEEPCNT."
            )
            self.raw_socket.setsockopt(
                socket.SOL_TCP,
                socket.TCP_KEEPIDLE,
                5 * 60
            )
            self.raw_socket.setsockopt(
                socket.IPPROTO_TCP,
                socket.TCP_KEEPIDLE,
                5 * 60
            )
            self.raw_socket.setsockopt(
                socket.IPPROTO_TCP,
                socket.TCP_KEEPINTVL,
                60
            )
            self.raw_socket.setsockopt(
                socket.IPPROTO_TCP,
                socket.TCP_KEEPCNT,
                2
            )

        if hasattr(socket, "TCP_USER_TIMEOUT"):
            self.log.debug(
                f"Setting TCP_USER_TIMEOUT to {self.socket_timeout}ms."
            )
            self.raw_socket.setsockopt(
                socket.IPPROTO_TCP,
                socket.TCP_USER_TIMEOUT,
                self.socket_timeout
            )

        self.socket = self._ssl_wrap()

    def _ssl_wrap(self) -> ssl.SSLSocket:
        """
        Wrap socket.

        Wraps the socket using the SSL protocol as configured in the SSL
        context, with hostname verification enabled.

        Returns:
            An SSL wrapped socket.
        """
        return self.ssl_context.wrap_socket(
            self.raw_socket,
            server_hostname=self.address
        )

    def send(
        self,
        data: Union[str, bytes]
    ) -> bool:
        """
        Send the str/Bytes to the server.

        If given string, it is encoded as 'uft-8' & send.
        UNSURE(MBK): Should the encoding be moved outside this class?

        Returns:
            True, if the data could be send else
            False.
        """
        if isinstance(data, str):
            data = data.encode('utf-8')

        try:
            self.socket.sendall(data)
        except ConnectionError:
            msg = "Get a ConnectionError, while trying to send"
            self.log.exception(msg)
            self.reconnect()
            return False
        except socket.timeout:
            msg = "Get a socket.timeout, while trying to send"
            self.log.exception(msg)
            # UNSURE: How do we hit this?
            return False
        except OSError:
            # UNSURE:
            msg = "Get a OSError, while trying to send"
            self.log.exception(msg)
            self.reconnect()
            return False
        except TimeoutError:
            msg = "Get a TimeoutError, while trying to send"
            self.log.exception(msg)
            # UNSURE: How do we hit this?
            return False
        except AttributeError:
            return False
        else:
            self.log.debug(f"Raw Data Send: {data!r}")
            return True

    def receive(self, parser: Callable[[bytes], Any]) -> Any:
        """
        Socket receive method.

        Method that handles receiving data from a socket. Capable of handling
        data chunks.

        Args:
            Callable: A parser, that returns the parsed data.
                      On Parsing Error, it should raise a
                      ValueError TypeError or any subClasses of those.
                      (Like 'JSONDecodeError' & 'pydantic.ValidationError' is)

        Returns:
            The "parser"'s output.
        """
        data = []
        while self.socket or not self.killed.is_set():
            try:
                data_chunk = self.socket.recv(self.RECEIVE_SIZE)
            except socket.timeout:
                # This happens every 2 Sec as set in self._socket_setup.
                continue
            except OSError as err:
                # UNSURE:
                if not self.socket:
                    # Socket Closed.
                    return
                if self.killed.is_set():
                    return
                self.log.warning(f"Receive -> OSError: {err}")
                self.reconnect()
                continue
            except TimeoutError:
                # UNSURE:
                self.log.exception("Receive -> Timeout")
                self.reconnect()
                continue
            if data_chunk == b'':
                self.log.debug("Server Closed socket.")
                self.reconnect()
            data.append(data_chunk)

            try:
                parsed_data = parser(b"".join(data))
            except ValueError as err:  # parentClass for JSONDecodeError.
                self.log.debug(f'Parsing Error: {err}')
                pass
            except TypeError as err:  # parentClass for pydantic.ValidationError
                self.log.debug(f'Parsing Error: {err}')
                pass
            else:
                self.log.debug(f"Raw Data Received: {data}")
                return parsed_data

    def connect(self) -> Optional[bool]:
        """
        Connect to the server.

        Attempts a connection to the server on the provided address and port.

        Returns:
            'True' if the connection was successful.
        """
        if self.killed.is_set():
            return False

        self._socket_setup()

        try:
            self.log.info("Trying to Connect.")
            self.observer.post(StatusID.CONNECTING, None)
            # self.socket.settimeout(10)  # Why?
            self.socket.connect((self.address, self.port))
            # self.socket.settimeout(None)  # Why?
            self.log.info(
                f"Connected on interface: {self.socket.getsockname()[0]}"
            )
            self.observer.post(StatusID.CONNECTED, None)
            # if self.sockt_thread is None:
            #     self._start()
            return True

        except Exception as e:
            self.log.error("Failed to connect: {}".format(e))
            raise

    def reconnect(self, retry_limit: Optional[int] = None) -> bool:
        """
        Attempt to reconnect.

        Reconnect to the server, until the given amount af attempts,
        are above the retry_limit.
        if the retry_limit are not set, it will never end.

        Returns:
            'True' if the connection was successful else
            'False'
        """
        if self.killed.is_set():
            return False

        if not self.socket:
            return False

        self.log.warning("Reconnection...")

        while retry_limit is None or retry_limit > 0:
            if retry_limit:
                retry_limit -= 1
            self.disconnect()
            try:
                if self.connect():
                    self.log.warning("Reconnected...")
                    return True
            except OSError:
                pass  # NOTE: Happens if it have forgotten the IP for the url.
            self.log.warning("Trying to reconnect in 5 seconds")
            time.sleep(5)
        return False

    def disconnect(self) -> None:
        """Disconnect from the server."""
        if self.socket:
            self.socket.close()

    def close(self) -> None:
        """
        Close the connection.

        Closes the socket object connection.
        """
        self.killed.set()
        self.log.info("Closing connection...")
        self.observer.post(StatusID.DISCONNECTING, None)
        if self.socket:
            self.socket.close()
            self.socket = None
        if self.raw_socket:
            self.raw_socket.close()
            self.raw_socket = None
        self.observer.post(StatusID.DISCONNETCED, None)
        self.log.info("Connection closed!")



"""Contain the Value Templates for Wappsto."""
from enum import Enum

from typing import Optional
from typing import Dict

from pydantic import BaseModel


# #############################################################################
#                             Value Settings Schema
# #############################################################################

class IoTEvent(str, Enum):
    """Different IoT Event."""
    CREATE = "create"  # POST
    CHANGE = "change"  # PUT
    REQUEST = "request"  # GET
    DELETE = "delete"  # DELETE


class ValueBaseType(str, Enum):
    """Internal use only!."""
    STRING = "string"
    NUMBER = "number"
    BLOB = "blob"
    XML = "xml"


class ValueSettinsSchema(BaseModel):
    """The Structure for which all templates should follow."""
    value_type: ValueBaseType
    type: str
    mapping: Optional[Dict[str, str]] = None  # Number only
    ordered_mapping: Optional[bool] = None  # Number only
    meaningful_zero: Optional[bool] = None  # Number only
    si_conversion: Optional[str] = None  # Number only
    min: Optional[str] = None  # Number only
    max: Optional[str] = None  # Blob, number, str only.
    step: Optional[str] = None  # Number only
    encoding: Optional[str] = None  # Blob, str only.
    xsd: Optional[str] = None  # XML only
    namespace: Optional[str] = None  # XML only
    unit: Optional[str] = None  # Number only


class ValueTemplate(str, Enum):
    """
    Predefined ValueTemplate.

    Each of the predefined ValueTemplate, have default
    value parameters set, which include BaseType, name,
    permission, range, step and the unit.
    """
    __version__ = "0.0.5"

    ADDRESS_NAME = "ADDRESS_NAME"
    ALTITUDE_M = "ALTITUDE_M"
    ANGLE = "ANGLE"
    APPARENT_POWER_VA = "APPARENT_POWER_VA"
    BLOB = "BLOB"
    BOOLEAN_ONOFF = "BOOLEAN_ONOFF"
    BOOLEAN_TRUEFALSE = "BOOLEAN_TRUEFALSE"
    CITY = "CITY"
    CO2_PPM = "CO2_PPM"
    COLOR_HEX = "COLOR_HEX"
    COLOR_INT = "COLOR_INT"
    COLOR_TEMPERATURE = "COLOR_TEMPERATURE"
    CONCENTRATION_PPM = "CONCENTRATION_PPM"
    CONNECTION_STATUS = "CONNECTION_STATUS"
    COUNT = "COUNT"
    COUNTRY = "COUNTRY"
    COUNTRY_CODE = "COUNTRY_CODE"
    CURRENT_A = "CURRENT_A"
    DISTANCE_M = "DISTANCE_M"
    DURATION_MIN = "DURATION_MIN"
    DURATION_MSEC = "DURATION_MSEC"
    DURATION_SEC = "DURATION_SEC"
    EMAIL = "EMAIL"
    ENERGY_KWH = "ENERGY_KWH"
    ENERGY_MWH = "ENERGY_MWH"
    ENERGY_PRICE_DKK_KWH = "ENERGY_PRICE_DKK_KWH"
    ENERGY_PRICE_DKK_MWH = "ENERGY_PRICE_DKK_MWH"
    ENERGY_PRICE_EUR_KWH = "ENERGY_PRICE_EUR_KWH"
    ENERGY_PRICE_EUR_MWH = "ENERGY_PRICE_EUR_MWH"
    ENERGY_WH = "ENERGY_WH"
    FREQUENCY_HZ = "FREQUENCY_HZ"
    HUMIDITY = "HUMIDITY"
    IDENTIFIER = "IDENTIFIER"
    IMAGE_JPG = "IMAGE_JPG"
    IMAGE_PNG = "IMAGE_PNG"
    IMPULSE_KWH = "IMPULSE_KWH"
    INTEGER = "INTEGER"
    JSON = "JSON"
    LATITUDE = "LATITUDE"
    LOAD_CURVE_ENERGY_KWH = "LOAD_CURVE_ENERGY_KWH"
    LOAD_CURVE_ENERGY_MWH = "LOAD_CURVE_ENERGY_MWH"
    LOAD_CURVE_ENERGY_WH = "LOAD_CURVE_ENERGY_WH"
    LONGITUDE = "LONGITUDE"
    LUMINOSITY_LX = "LUMINOSITY_LX"
    NUMBER = "NUMBER"
    ORGANIZATION = "ORGANIZATION"
    PERCENTAGE = "PERCENTAGE"
    PHONE = "PHONE"
    POSTCODE = "POSTCODE"
    POWER_KW = "POWER_KW"
    POWER_WATT = "POWER_WATT"
    PRECIPITATION_MM = "PRECIPITATION_MM"
    PRESSURE_HPA = "PRESSURE_HPA"
    REACTIVE_ENERGY_KVARH = "REACTIVE_ENERGY_KVARH"
    REACTIVE_POWER_KVAR = "REACTIVE_POWER_KVAR"
    SPEED_KMH = "SPEED_KMH"
    SPEED_MS = "SPEED_MS"
    STREET = "STREET"
    STRING = "STRING"
    TEMPERATURE_CELSIUS = "TEMPERATURE_CELSIUS"
    TEMPERATURE_FAHRENHEIT = "TEMPERATURE_FAHRENHEIT"
    TEMPERATURE_KELVIN = "TEMPERATURE_KELVIN"
    TIMESTAMP = "TIMESTAMP"
    TIME_OF_DAY = "TIME_OF_DAY"
    TOTAL_ENERGY_KWH = "TOTAL_ENERGY_KWH"
    TOTAL_ENERGY_MWH = "TOTAL_ENERGY_MWH"
    TOTAL_ENERGY_WH = "TOTAL_ENERGY_WH"
    TRIGGER = "TRIGGER"
    UNIT_TIME = "UNIT_TIME"
    VOLTAGE_V = "VOLTAGE_V"
    VOLUME_M3 = "VOLUME_M3"
    XML = "XML"


valueSettings: Dict[ValueTemplate, ValueSettinsSchema] = {

    ValueTemplate.TRIGGER: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="trigger",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="0",
        step="0",
        unit=None,
        si_conversion=None,
    ),
    ValueTemplate.BOOLEAN_TRUEFALSE: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="boolean",
        mapping={'0': 'false', '1': 'true'},
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="1",
        step="1",
        unit=None,
        si_conversion=None,
    ),
    ValueTemplate.BOOLEAN_ONOFF: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="boolean",
        mapping={'0': 'off', '1': 'on'},
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="1",
        step="1",
        unit=None,
        si_conversion=None,
    ),
    ValueTemplate.CONNECTION_STATUS: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="connection",
        mapping={'0': 'offline', '1': 'online'},
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="1",
        step="1",
        unit=None,
        si_conversion=None,
    ),
    ValueTemplate.INTEGER: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="integer",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="-255",
        max="255",
        step="1",
        unit=None,
        si_conversion=None,
    ),
    ValueTemplate.COUNT: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="count",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="255",
        step="1",
        unit=None,
        si_conversion=None,
    ),
    ValueTemplate.IMPULSE_KWH: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="impulse_resolution",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=True,
        min="1",
        max="50000",
        step="1",
        unit="imp/kWh",
        si_conversion=None,
    ),
    ValueTemplate.VOLTAGE_V: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="voltage",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="250",
        step="0.1",
        unit="V",
        si_conversion=None,
    ),
    ValueTemplate.POWER_WATT: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="power",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="-1000000",
        max="2500",
        step="0.1",
        unit="W",
        si_conversion=None,
    ),
    ValueTemplate.POWER_KW: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="power",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="-1000000",
        max="1000000",
        step="0.1",
        unit="kW",
        si_conversion="[W] = 1000 * [kW]",
    ),
    ValueTemplate.ENERGY_WH: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="energy",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="-1000000",
        max="100000",
        step="0.1",
        unit="Wh",
        si_conversion="[J] = 3600 * [Wh]",
    ),
    ValueTemplate.ENERGY_KWH: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="energy",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="-1000000",
        max="1000000",
        step="0.1",
        unit="kWh",
        si_conversion="[J] = 3600000 * [kWh]",
    ),
    ValueTemplate.ENERGY_MWH: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="energy",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="-1000000",
        max="1000000",
        step="0.1",
        unit="MWh",
        si_conversion="[J] = 3600000000 * [MWh]",
    ),
    ValueTemplate.TOTAL_ENERGY_WH: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="total_energy",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="1000000",
        step="0.1",
        unit="Wh",
        si_conversion="[J] = 3600 * [Wh]",
    ),
    ValueTemplate.TOTAL_ENERGY_KWH: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="total_energy",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="1000000",
        step="0.1",
        unit="kWh",
        si_conversion="[J] = 3600000 * [kWh]  ",
    ),
    ValueTemplate.TOTAL_ENERGY_MWH: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="total_energy",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="1000000",
        step="0.1",
        unit="MWh",
        si_conversion="[J] = 3600000000 * [MWh]",
    ),
    ValueTemplate.LOAD_CURVE_ENERGY_WH: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="load_curve_energy",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="-1000000",
        max="1000000",
        step="0.1",
        unit="Wh",
        si_conversion="[J] = 3600 * [Wh]",
    ),
    ValueTemplate.LOAD_CURVE_ENERGY_KWH: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="load_curve_energy",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="-1000000",
        max="1000000",
        step="0.1",
        unit="kWh",
        si_conversion="[J] = 3600000 * [kWh]  ",
    ),
    ValueTemplate.LOAD_CURVE_ENERGY_MWH: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="load_curve_energy",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="-1000000",
        max="1000000",
        step="0.1",
        unit="MWh",
        si_conversion="[J] = 3600000000 * [MWh]",
    ),
    ValueTemplate.CURRENT_A: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="electric_current",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="-5000",
        max="5000",
        step="0.001",
        unit="A",
        si_conversion=None,
    ),
    ValueTemplate.APPARENT_POWER_VA: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="apparent_power",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="-5000",
        max="5000",
        step="0.001",
        unit="VA",
        si_conversion=None,
    ),
    ValueTemplate.REACTIVE_ENERGY_KVARH: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="reactive_energy",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="-5000",
        max="5000",
        step="0.001",
        unit="kvarh",
        si_conversion=None,
    ),
    ValueTemplate.REACTIVE_POWER_KVAR: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="reactive_power",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="-5000",
        max="5000",
        step="0.001",
        unit="kvar",
        si_conversion=None,
    ),
    ValueTemplate.ENERGY_PRICE_EUR_KWH: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="energy_price",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=False,
        min="-10000",
        max="10000",
        step="0.01",
        unit="EUR/kWh",
        si_conversion=None,
    ),
    ValueTemplate.ENERGY_PRICE_EUR_MWH: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="energy_price",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=False,
        min="-10000",
        max="10000",
        step="0.001",
        unit="EUR/MWh",
        si_conversion=None,
    ),
    ValueTemplate.ENERGY_PRICE_DKK_KWH: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="energy_price",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=False,
        min="-10000",
        max="10000",
        step="0.01",
        unit="DKK/kWh",
        si_conversion=None,
    ),
    ValueTemplate.ENERGY_PRICE_DKK_MWH: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="energy_price",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=False,
        min="-10000",
        max="10000",
        step="0.001",
        unit="DKK/MWh",
        si_conversion=None,
    ),
    ValueTemplate.FREQUENCY_HZ: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="frequency",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="30000",
        step="0.01",
        unit="Hz",
        si_conversion=None,
    ),
    ValueTemplate.TEMPERATURE_CELSIUS: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="temperature",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=False,
        min="-30",
        max="50",
        step="1",
        unit="°C",
        si_conversion="[K] = [°C] + 273.15",
    ),
    ValueTemplate.TEMPERATURE_FAHRENHEIT: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="temperature",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=False,
        min="-20",
        max="120",
        step="1",
        unit="°F",
        si_conversion="[K] = ([°F] + 459.67) × 5/9 ",
    ),
    ValueTemplate.TEMPERATURE_KELVIN: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="temperature",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=True,
        min="240",
        max="320",
        step="1",
        unit="K",
        si_conversion=None,
    ),
    ValueTemplate.ANGLE: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="angle",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="360",
        step="0",
        unit="°",
        si_conversion="[rad] = (180/pi) * [°]",
    ),
    ValueTemplate.PERCENTAGE: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="percentage",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="100",
        step="1",
        unit="%",
        si_conversion="[1] = 100 * [%]",
    ),
    ValueTemplate.SPEED_MS: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="speed",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="100",
        step="1",
        unit="m/s",
        si_conversion=None,
    ),
    ValueTemplate.SPEED_KMH: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="speed",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=True,
        min="0",
        max="400",
        step="0.1",
        unit="km/h",
        si_conversion="[ms] = [kmh]*1000/3600",
    ),
    ValueTemplate.PRECIPITATION_MM: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="precipitation",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="100",
        step="1",
        unit="mm",
        si_conversion=None,
    ),
    ValueTemplate.HUMIDITY: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="relative_humidity",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=True,
        min="0",
        max="100",
        step="1",
        unit="%",
        si_conversion="[1] = 100 * [%]",
    ),
    ValueTemplate.CO2_PPM: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="co2",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=True,
        min="0",
        max="3000",
        step="1",
        unit="ppm",
        si_conversion="1000000 * [ppm]",
    ),
    ValueTemplate.CONCENTRATION_PPM: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="concentration",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=True,
        min="0",
        max="3000",
        step="1",
        unit="ppm",
        si_conversion="1000000 * [ppm]",
    ),
    ValueTemplate.PRESSURE_HPA: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="pressure",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="300",
        max="1100",
        step="1",
        unit="hPa",
        si_conversion="[Pa] = [hPa]/100",
    ),
    ValueTemplate.VOLUME_M3: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="volume",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=True,
        min="0",
        max="1000000000",
        step="0.001",
        unit="m³",
        si_conversion="[m³] = [m³]",
    ),
    ValueTemplate.UNIT_TIME: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="timestamp",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=True,
        min="0",
        max="2147483647",
        step="1",
        unit="s",
        si_conversion="[s] = [s]",
    ),
    ValueTemplate.TIMESTAMP: ValueSettinsSchema(
        value_type=ValueBaseType.STRING,
        type="timestamp",
        max="27",
        encoding="ISO 8601",
    ),
    ValueTemplate.DURATION_MIN: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="duration",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="1440",
        step="0.1",
        unit="min",
        si_conversion="[s] = [min] / 60",
    ),
    ValueTemplate.DURATION_SEC: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="duration",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="3600",
        step="0.001",
        unit="s",
        si_conversion="[s] = [s]",
    ),
    ValueTemplate.DURATION_MSEC: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="duration",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="5000",
        step="0.001",
        unit="ms",
        si_conversion="[s] = [ms]/1000",
    ),
    ValueTemplate.TIME_OF_DAY: ValueSettinsSchema(
        value_type=ValueBaseType.STRING,
        type="time",
        max="100",
        encoding="",
    ),
    ValueTemplate.DISTANCE_M: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="distance",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="1000",
        step="1",
        unit="m",
        si_conversion=None,
    ),
    ValueTemplate.LUMINOSITY_LX: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="luminosity",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="25000",
        step="1",
        unit="lx",
        si_conversion=None,
    ),
    ValueTemplate.COLOR_HEX: ValueSettinsSchema(
        value_type=ValueBaseType.BLOB,
        type="color",
        max="6",
        encoding="hex",
    ),
    ValueTemplate.COLOR_INT: ValueSettinsSchema(
        value_type=ValueBaseType.BLOB,
        type="color",
        max="8",
        encoding="integer",
    ),
    ValueTemplate.COLOR_TEMPERATURE: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="color_temperature",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="1000",
        max="12000",
        step="1",
        unit="K",
        si_conversion=None,
    ),
    ValueTemplate.IMAGE_JPG: ValueSettinsSchema(
        value_type=ValueBaseType.BLOB,
        type="image",
        max="10485100",
        encoding="base64;jpg",
    ),
    ValueTemplate.IMAGE_PNG: ValueSettinsSchema(
        value_type=ValueBaseType.BLOB,
        type="image",
        max="10485100",
        encoding="base64;png",
    ),
    ValueTemplate.LATITUDE: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="latitude",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="-90",
        max="90",
        step="0.000001",
        unit="°N",
        si_conversion=None,
    ),
    ValueTemplate.LONGITUDE: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="longitude",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="-180",
        max="180",
        step="0.000001",
        unit="°E",
        si_conversion=None,
    ),
    ValueTemplate.ALTITUDE_M: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="altitude",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="-10000",
        max="10000",
        step="0.01",
        unit="m",
        si_conversion=None,
    ),
    ValueTemplate.STREET: ValueSettinsSchema(
        value_type=ValueBaseType.STRING,
        type="street",
        max="85",
        encoding="",
    ),
    ValueTemplate.CITY: ValueSettinsSchema(
        value_type=ValueBaseType.STRING,
        type="city",
        max="85",
        encoding="",
    ),
    ValueTemplate.POSTCODE: ValueSettinsSchema(
        value_type=ValueBaseType.STRING,
        type="postcode",
        max="10",
        encoding="",
    ),
    ValueTemplate.COUNTRY: ValueSettinsSchema(
        value_type=ValueBaseType.STRING,
        type="country",
        max="56",
        encoding="",
    ),
    ValueTemplate.COUNTRY_CODE: ValueSettinsSchema(
        value_type=ValueBaseType.STRING,
        type="country_code",
        max="2",
        encoding="ISO 3166-1 Alpha-2",
    ),
    ValueTemplate.ADDRESS_NAME: ValueSettinsSchema(
        value_type=ValueBaseType.STRING,
        type="address_name",
        max="85",
        encoding="",
    ),
    ValueTemplate.ORGANIZATION: ValueSettinsSchema(
        value_type=ValueBaseType.STRING,
        type="organization",
        max="85",
        encoding="",
    ),
    ValueTemplate.EMAIL: ValueSettinsSchema(
        value_type=ValueBaseType.STRING,
        type="email",
        max="128",
        encoding="",
    ),
    ValueTemplate.PHONE: ValueSettinsSchema(
        value_type=ValueBaseType.STRING,
        type="phone",
        max="32",
        encoding="",
    ),
    ValueTemplate.IDENTIFIER: ValueSettinsSchema(
        value_type=ValueBaseType.STRING,
        type="identifier",
        max="50",
        encoding="",
    ),
    ValueTemplate.JSON: ValueSettinsSchema(
        value_type=ValueBaseType.BLOB,
        type="json",
        max="20000",
        encoding="json",
    ),
    ValueTemplate.NUMBER: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="number",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="-128",
        max="128",
        step="0.1",
        unit=None,
        si_conversion=None,
    ),
    ValueTemplate.STRING: ValueSettinsSchema(
        value_type=ValueBaseType.STRING,
        type="string",
        max="64",
        encoding="",
    ),
    ValueTemplate.BLOB: ValueSettinsSchema(
        value_type=ValueBaseType.BLOB,
        type="blob",
        max="280",
        encoding="base64",
    ),
    ValueTemplate.XML: ValueSettinsSchema(
        value_type=ValueBaseType.XML,
        type="xml",
        xsd="",
        namespace="",
    ),
}



"""Contain the basic Wappsto schema for the network structure and children."""
from datetime import datetime
from enum import Enum

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
# from typing import TypeAlias
from typing import Union

from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import conint
# from pydantic import constr
from pydantic import Field
from pydantic import field_serializer
from pydantic import GenerateSchema
from pydantic import UUID4

from pydantic_core import CoreSchema
from pydantic_core import core_schema


class LaxStrGenerator(GenerateSchema):
    """
    A more relax converter, that make it work like Pydantic v1.

    URL: https://github.com/pydantic/pydantic/issues/6045#issuecomment-1650443311
    """

    def str_schema(self) -> CoreSchema:
        """Convert a string more relaxed."""
        return core_schema.no_info_before_validator_function(str, core_schema.str_schema())


BaseModel.model_config = ConfigDict(schema_generator=LaxStrGenerator)


def timestamp_converter(dt: datetime) -> Optional[str]:
    """
    Return The default timestamp used for Wappsto.

    The timestamp are always set to the UTC timezone.

    Returns:
        The UTC time string in ISO format.
    """
    if dt is None:
        return None
    return dt.strftime('%Y-%m-%dT%H:%M:%S.%fZ')


class WappstoMethods(str, Enum):
    """The methods that the Wappsto are using."""

    DELETE = "DELETE"
    PUT = "PUT"
    POST = "POST"
    GET = "GET"


class InclusionStatus(str, Enum):
    """The possible inclusion status for a device."""

    STATUS_DEVICE_INCLUDING = 'STATUS_DEVICE_INCLUDING'
    STATUS_DEVICE_INCLUSION_SUCCESS = 'STATUS_DEVICE_INCLUSION_SUCCESS'
    STATUS_DEVICE_INCLUSION_FAILURE = 'STATUS_DEVICE_INCLUSION_FAILURE'
    STATUS_DEVICE_REPORTING = 'STATUS_DEVICE_REPORTING'
    STATUS_DEVICE_REPORT_SUCCESS = 'STATUS_DEVICE_REPORT_SUCCESS'
    STATUS_DEVICE_REPORT_FAILURE = 'STATUS_DEVICE_REPORT_FAILURE'
    EXCLUDED = 'EXCLUDED'
    INCLUDED = 'INCLUDED'


class FirmwareStatus(str, Enum):
    """The possible firmware status for a device."""

    UP_TO_DATE = 'UP_TO_DATE'
    UPDATE_AVAILABLE = 'UPDATE_AVAILABLE'
    UPLOADING = 'UPLOADING'
    UPLOAD_COMPLETE = 'UPLOAD_COMPLETE'
    UPLOADING_FAILURE = 'UPLOADING_FAILURE'
    FLASHING = 'FLASHING'
    FLASHING_COMPLETE = 'FLASHING_COMPLETE'
    FLASHING_FAILURE = 'FLASHING_FAILURE'


class Command(str, Enum):
    """The possible firmware status for a device."""

    IDLE = 'idle'
    FIRMWARE_UPLOAD = 'firmware_upload'
    FIRMWARE_FLASH = 'firmware_flash'
    FIRMWARE_CANCEL = 'firmware_cancel'
    INCLUDE = 'include'
    EXCLUDE = 'exclude'
    CONNECTION_CHECK = 'connection_check'


class OwnerEnum(str, Enum):
    """The Owner of the Object."""

    UNASSIGNED = 'unassigned'


class Deletion(str, Enum):
    """The Status of a deletion request."""

    PENDING = 'pending'
    FAILED = 'failed'


class WappstoVersion(str, Enum):
    """The different version available."""

    V2_0 = "2.0"
    V2_1 = "2.1"


class PermissionType(str, Enum):
    """All possible Value Permission Types."""

    READ = 'r'
    WRITE = 'w'
    READWRITE = 'rw'
    WRITEREAD = 'wr'
    NONE = 'none'


class EventStatus(str, Enum):
    """Update status/request."""

    OK = 'ok'
    UPDATE = 'update'
    PENDING = 'pending'


class StateType(str, Enum):
    """All the State type."""

    REPORT = 'Report'
    CONTROL = 'Control'


class StateStatus(str, Enum):
    """Update status/request."""

    SEND = 'Send'
    PENDING = 'Pending'
    FAILED = 'Failed'


class Level(str, Enum):
    """The different log levels."""

    IMPORTANT = 'important'
    ERROR = 'error'
    WARNING = 'warning'
    SUCCESS = 'success'
    INFO = 'info'
    DEBUG = 'debug'


class StatusType(str, Enum):
    """The different status types."""

    PUBLIC_KEY = 'public key'
    MEMORY_INFORMATION = 'memory information'
    DEVICE_DESCRIPTION = 'device description'
    VALUE_DESCRIPTION = 'value description'
    VALUE = 'value'
    PARTNER_INFORMATION = 'partner information'
    ACTION = 'action'
    CALCULATION = 'calculation'
    TIMER = 'timer'
    CALENDAR = 'calendar'
    STATEMACHINE = 'statemachine'
    FIRMWARE_UPDATE = 'firmware update'
    CONFIGURATION = 'configuration'
    EXI = 'exi'
    SYSTEM = 'system'
    APPLICATION = 'application'
    GATEWAY = 'gateway'


class WappstoMetaType(str, Enum):
    """
    All possible Wappsto Meta Types.

    They have a parent child relation in order of:
    network->device->value->state

    Where a 'Network' only contains 'Device's,
    'Device's only contains 'value's, and
    'Value's only contains 'State's.
    """

    NETWORK = "network"
    DEVICE = "device"
    VALUE = "value"
    STATE = "state"
    CREATOR = "creator"
    IDLIST = "idlist"
    DELETELIST = "deletelist"


class Connection(BaseModel):
    """The Connection info for the network."""

    timestamp: Optional[datetime] = None
    online: Optional[bool] = None

    @field_serializer('timestamp')
    def serialize_timestamp(self, value: datetime) -> Optional[str]:
        """Convert datetime to Wappsto datetime standard."""
        return timestamp_converter(value)


class WarningItem(BaseModel):
    """The Connection info for the network."""

    message: Optional[Optional[str]] = None
    data: Optional[Optional[Dict[str, Any]]] = None
    code: Optional[Optional[int]] = None


class Geo(BaseModel):
    """The geolocation structure for network & device objects."""

    latitude: Optional[str] = None
    longitude: Optional[str] = None
    display_name: Optional[str] = None
    address: Optional[Dict[str, Any]] = None


class BaseMeta(BaseModel):
    """The base for all meta objects."""

    id: Optional[UUID4] = None
    # NOTE: Set in the children-class to enforce right BaseModel type.
    # #  type: Optional[WappstoMetaType] = None
    version: Optional[WappstoVersion] = None

    manufacturer: Optional[UUID4] = None
    owner: Optional[Union[UUID4, OwnerEnum]] = None
    parent: Optional[UUID4] = None

    created: Optional[datetime] = None
    updated: Optional[datetime] = None
    changed: Optional[datetime] = None

    application: Optional[UUID4] = None
    deletion: Optional[Deletion] = None
    deprecated: Optional[Optional[bool]] = None

    iot: Optional[Optional[bool]] = None
    revision: Optional[Optional[int]] = None
    size: Optional[Optional[int]] = None
    path: Optional[Optional[str]] = None

    oem: Optional[Optional[str]] = None
    accept_manufacturer_as_owner: Optional[Optional[bool]] = None

    # NOTE: patterns do not work while the `LaxStrGenerator` are in use.
    # redirect: Optional[  # type: ignore
    #     constr(
    #         pattern=r'^[0-9a-zA-Z_-]+$',  # noqa: F722
    #         min_length=1,
    #         max_length=200
    #     )
    # ] = None

    error: Optional[UUID4] = None
    warning: Optional[List[WarningItem]] = None
    trace: Optional[Optional[str]] = None

    set: Optional[List[UUID4]] = None
    contract: Optional[List[UUID4]] = None

    historical: Optional[bool] = None

    @field_serializer('created', 'updated', 'changed')
    def serialize_timestamp(self, value: datetime) -> Optional[str]:
        """Convert datetime to Wappsto datetime standard."""
        return timestamp_converter(value)


class EventlogMeta(BaseMeta):
    """The meta object for the Eventlog structure."""

    class WappstoMetaType(str, Enum):
        """Used instead of typing.Literal."""

        STATUS = "eventlog"

    type: Optional[WappstoMetaType] = None

    icon: Optional[Optional[str]] = None
    alert: Optional[List[UUID4]] = None


class StatusMeta(BaseMeta):
    """The meta object for the Status structure."""

    class WappstoMetaType(str, Enum):
        """Used instead of typing.Literal."""

        STATUS = "status"
    type: Optional[WappstoMetaType] = None

    icon: Optional[Optional[str]] = None
    alert: Optional[List[UUID4]] = None


class ValueMeta(BaseMeta):
    """The meta object for the Value structure."""

    class WappstoMetaType(str, Enum):
        """Used instead of typing.Literal."""

        VALUE = "value"
    type: Optional[WappstoMetaType] = None


class StateMeta(BaseMeta):
    """The meta object for the State structure."""

    class WappstoMetaType(str, Enum):
        """Used instead of typing.Literal."""

        STATE = "state"
    type: Optional[WappstoMetaType] = None


class DeviceMeta(BaseMeta):
    """The meta object for the Device structure."""

    class WappstoMetaType(str, Enum):
        """Used instead of typing.Literal."""

        DEVICE = "device"

    type: Optional[WappstoMetaType] = None

    geo: Optional[Geo] = None


class NetworkMeta(BaseMeta):
    """The meta object for the Network structure."""

    class WappstoMetaType(str, Enum):
        """Used instead of typing.Literal."""

        NETWORK = "network"

    type: Optional[WappstoMetaType] = None

    geo: Optional[Geo] = None
    connection: Optional[Connection] = None
    accept_test_mode: Optional[bool] = None
    verify_product: Optional[str] = None
    product: Optional[str] = None


class Status(BaseModel):
    """Contain the status."""

    message: str
    timestamp: datetime
    data: Optional[str] = None
    level: Level
    type: Optional[StatusType]
    meta: Optional[StatusMeta] = Field(None, title='meta-2.0:create')

    @field_serializer('timestamp')
    def serialize_timestamp(self, value: datetime) -> Optional[str]:
        """Convert datetime to Wappsto datetime standard."""
        return timestamp_converter(value)


class Info(BaseModel):
    """."""

    enabled: Optional[bool] = None


class LogValue(BaseModel):
    """The required data for post of new values."""

    data: str
    timestamp: datetime

    model_config: ConfigDict = ConfigDict(extra='forbid')  # type: ignore

    @field_serializer('timestamp')
    def serialize_timestamp(self, value: datetime) -> Optional[str]:
        """Convert datetime to Wappsto datetime standard."""
        return timestamp_converter(value)


class State(BaseModel):
    """The State object found in values, that contain the raw value."""

    data: str
    type: Optional[StateType] = None

    meta: Optional[StateMeta] = Field(None, title='meta-2.0:create')
    status: Optional[StateStatus] = None
    status_payment: Optional[str] = None
    timestamp: Optional[datetime] = None

    model_config: ConfigDict = ConfigDict(extra='forbid')  # type: ignore

    @field_serializer('timestamp')
    def serialize_timestamp(self, value: datetime) -> Optional[str]:
        """Convert datetime to Wappsto datetime standard."""
        return timestamp_converter(value)


class EventlogItem(BaseModel):
    """Event Log structure, found in values.."""

    message: str
    timestamp: Optional[datetime] = None
    info: Optional[Dict[str, Any]] = None
    level: Level
    type: Optional[str] = None
    meta: Optional[EventlogMeta] = Field(None, title='meta-2.0:create')

    @field_serializer('timestamp')
    def serialize_timestamp(self, value: datetime) -> Optional[str]:
        """Convert datetime to Wappsto datetime standard."""
        return timestamp_converter(value)


class BaseValue(BaseModel):
    """Base structure for all values types, what all values have."""

    name: Optional[str] = None
    type: Optional[str] = None
    description: Optional[str] = None
    period: Optional[str] = None
    delta: Optional[str] = None
    permission: Optional[PermissionType] = None
    status: Optional[EventStatus] = None
    meta: Optional[ValueMeta] = Field(None, title='meta-2.0:create')
    state: Optional[List[Union[State, UUID4]]] = None
    eventlog: Optional[List[Union[EventlogItem, UUID4]]] = None
    info: Optional[Info] = None

    model_config: ConfigDict = ConfigDict(extra='forbid')  # type: ignore


class Number(BaseModel):
    """Substructure for the Number value."""

    min: Union[float, int, str]
    max: Union[float, int, str]
    step: Union[float, int, str]
    mapping: Optional[Dict[str, Any]] = None
    meaningful_zero: Optional[bool] = None
    ordered_mapping: Optional[bool] = None
    si_conversion: Optional[str] = None
    unit: Optional[str] = None


class String(BaseModel):
    """Substructure for the String value."""

    max: Optional[conint(ge=1, multiple_of=1)] = None  # type: ignore
    encoding: Optional[str] = None


class Blob(BaseModel):
    """Substructure for the Blob value."""

    max: Optional[conint(ge=1, multiple_of=1)] = None  # type: ignore
    encoding: Optional[str] = None


class Xml(BaseModel):
    """Substructure for the XML value."""

    xsd: Optional[str] = None
    namespace: Optional[str] = None


class StringValue(BaseValue):
    """One of four Value type."""

    string: Optional[String] = None

    model_config: ConfigDict = ConfigDict(extra='forbid')  # type: ignore

    # @model_validator(mode='before')
    # def value_type_check(cls, values: Dict[str, Any]):
    #     """Force the Validate for the value type."""
    #     keys = ["number", "blob", "xml"]
    #     if any(key in values for key in keys):
    #         raise TypeError(f"A Wappsto string value can not contain: {' '.join(keys)}")
    #     return values


class NumberValue(BaseValue):
    """One of four Value type."""

    number: Optional[Number] = None

    model_config: ConfigDict = ConfigDict(extra='forbid')  # type: ignore

    # @model_validator(mode='before')
    # def value_type_check(cls, values: Dict[str, Any]):
    #     """Force the Validate for the value type."""
    #     keys = ["string", "blob", "xml"]
    #     if any(key in values for key in keys):
    #         raise TypeError(f"A Wappsto number value can not contain: {' '.join(keys)}")
    #     return values


class BlobValue(BaseValue):
    """One of four Value type."""

    blob: Optional[Blob] = None

    model_config: ConfigDict = ConfigDict(extra='forbid')  # type: ignore

    # @model_validator(mode='before')
    # def value_type_check(cls, values: Dict[str, Any]):
    #     """Force the Validate for the value type."""
    #     keys = ["number", "string", "xml"]
    #     if any(key in values for key in keys):
    #         raise TypeError(f"A Wappsto blob value can not contain: {' '.join(keys)}")
    #     return values


class XmlValue(BaseValue):
    """One of four Value type."""

    xml: Optional[Xml] = None

    model_config: ConfigDict = ConfigDict(extra='forbid')  # type: ignore

    # @model_validator(mode='before')
    # def value_type_check(cls, values: Dict[str, Any]):
    #     """Force the Validate for the value type."""
    #     keys = ["number", "blob", "string"]
    #     if any(key in values for key in keys):
    #         raise TypeError(f"A Wappsto xml value can not contain: {' '.join(keys)}")
    #     return values


class Device(BaseModel):
    """The Wappsto device structure."""

    name: Optional[str] = None
    control_timeout: Optional[int] = None
    control_when_offline: Optional[bool] = None
    manufacturer: Optional[str] = None
    product: Optional[str] = None
    version: Optional[str] = None
    serial: Optional[str] = None
    description: Optional[str] = None
    protocol: Optional[str] = None
    communication: Optional[str] = None
    included: Optional[str] = None
    inclusion_status: Optional[InclusionStatus] = None
    firmware_status: Optional[FirmwareStatus] = None
    firmware_upload_progress: Optional[str] = None
    firmware_available_version: Optional[str] = None
    command: Optional[Command] = None
    meta: Optional[DeviceMeta] = Field(None, title='meta-2.0:create')
    status: Optional[List[Union[Status, UUID4]]] = None
    value: Optional[
        List[
            Union[
                StringValue,
                NumberValue,
                BlobValue,
                XmlValue,
                UUID4
            ]
        ]
    ] = None
    info: Optional[Info] = None

    # model_config: ConfigDict = ConfigDict(extra='forbid')  # type: ignore


class Network(BaseModel):
    """The root wappsto structure, for all IoT data."""

    name: Optional[str] = None
    description: Optional[str] = None
    device: Optional[List[Union[Device, UUID4]]] = None
    meta: Optional[NetworkMeta] = Field(None, title='meta-2.0:create')
    info: Optional[Info] = None

    # model_config: ConfigDict = ConfigDict(extra='forbid')  # type: ignore


class ApiMetaTypes(str, Enum):
    """The enum list of the different list types."""

    idlist = "idlist"
    deletelist = "deletelist"


class ApiMetaInfo(BaseModel):
    """The meta structure for list types."""

    type: ApiMetaTypes  # Merge with MetaAPIData?
    version: WappstoVersion


class childInfo(BaseModel):
    """The info of the returned object in the Id List."""

    type: WappstoMetaType
    version: WappstoVersion


class IdList(BaseModel):
    """The structure reply when a list of objects have been requested."""

    child: List[childInfo]
    id: List[UUID4]
    more: bool
    limit: int
    count: int
    meta: ApiMetaInfo

    model_config: ConfigDict = ConfigDict(extra='forbid')  # type: ignore


class DeleteList(BaseModel):
    """The structure reply when a delete request and been send."""

    deleted: List[UUID4]
    code: int
    message: str = "Deleted"
    meta: ApiMetaInfo


"""A collection of all Wappsto Value Types."""
Value = Union[StringValue, NumberValue, BlobValue, XmlValue]
ValueUnion = Union[StringValue, NumberValue, BlobValue, XmlValue]

"""A collection of all Wappsto Types."""
WappstoObject = Union[Network, Device, Value, State, IdList, DeleteList]



#!/usr/bin/env python3
import datetime
import pathlib
import pytest
import uuid
import wappstoiot

from typing import Optional

from utils import file_utils

# TODO: Fix this test.
# from wappstoiot.__main__ import _start_session
# from wappstoiot.__main__ import _create_network
# from wappstoiot.__main__ import _claim_network
# from wappstoiot.__main__ import _create_certificaties_files


def pytest_generate_tests(metafunc):
    """Attach the cmd-line args to a test-class that needs them."""
    token = metafunc.config.getoption("token")
    if token and hasattr(metafunc.cls, 'token'):
        metafunc.cls.token = token


class BaseTestClassWithCertificateFiles:
    """Contain the Base functionalities for a test with certificate files."""
    temp: pathlib.Path = pathlib.Path(__file__).parent.parent / pathlib.Path('temp')

    def generate_certificates(
        self,
        username: Optional[str] = None,
        password: Optional[str] = None,
        token: Optional[str] = None,
    ):
        """Generate the needed certificates for a 'real' world test."""
        base_url: str = "qa.wappsto.com"
        if not token:
            session = _start_session(
                base_url=base_url,
                username=username,
                password=password,
            )
        else:
            session = token

        creator = _create_network(
            session=session,
            base_url=base_url,
            dry_run=False
        )
        _claim_network(
            session=session,
            base_url=base_url,
            network_uuid=creator.get('network', {}).get('id'),
            dry_run=False
        )

        self.temp.mkdir(exist_ok=True)

        _create_certificaties_files(self.temp, creator, dry_run=False)


class TestWithOutMocking(BaseTestClassWithCertificateFiles):

    token: Optional[uuid.UUID] = None

    @classmethod
    def setup_class(cls):
        """Setting up all the needed stuff for the test class."""
        if cls.token is None:
            pytest.skip("Token was not given.")
        cls.generate_certificates(cls, token=cls.token)

    @classmethod
    def teardown_class(cls):
        """tearing down all the generated stuff by the setup_class function."""
        file_utils.rm_all(cls.temp)

    @pytest.mark.parametrize(
        "fast_send",
        [
            True,
            False,
        ]
    )
    def test_pytest_usage(self, fast_send: bool):
        # TODO: Hock into the observer, to check for communication error.
        # TODO: Go through the Rest-API to check if value was set.
        try:
            wappstoiot.config(
                config_folder=self.temp,
                fast_send=fast_send
            )
            network = wappstoiot.createNetwork("TheNetwork")
            device = network.createDevice("TheDevice")
            value = device.createValue(
                "TheValue",
                permission=wappstoiot.PermissionType.READ,
                value_template=wappstoiot.ValueTemplate.STRING
            )
            value.report(value=5)
            value.report(value=1337, timestamp=datetime.datetime.now())
        finally:
            wappstoiot.close()



"""Contain the basic JSONRpc structure for the the IoT endpoint."""
import uuid
import datetime

from enum import Enum
from itertools import zip_longest

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union
from typing import Iterable

from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import field_validator
from pydantic import FieldValidationInfo
from pydantic import TypeAdapter

from .base_schema import BlobValue
from .base_schema import Device
from .base_schema import Network
from .base_schema import NumberValue
from .base_schema import State
from .base_schema import LogValue
from .base_schema import StringValue
from .base_schema import XmlValue
from .base_schema import IdList
from .base_schema import DeleteList


def pair_wise(values: Iterable[str]) -> Iterable[Tuple[str, str]]:
    """Pair up the given values, two by two (hands of blue)."""
    a = iter(values)
    return zip_longest(a, a)


ValueUnion: Type = Union[StringValue, NumberValue, BlobValue, XmlValue]


JsonRpc_error_codes = {
    # Rpc Error Code: [HTTP Error Code, "Error String"]
    -32700: [400, "Parse error"],
    -32600: [400, "Invalid Request"],
    -32601: [404, "Method not found"],
    -32602: [400, "Invalid params"],
    -32603: [500, "Internal Server Error"],

    -32000: [404, "Timeout"],
    -32001: [301, "Moved Permanently"],
    -32002: [400, "Bad Request"],
    -32003: [401, "Unauthorized"],
    -32004: [402, "Payment Required"],
    -32005: [403, "Forbidden"],
    -32006: [404, "Not Found"],
    -32007: [405, "Method Not Allowed"],
    -32008: [406, "Not Acceptable"],
    -32009: [408, "Request Timeout"],
    -32010: [409, "Conflict"],
    -32011: [410, "Gone"],
    -32012: [415, "Unsupported Media Type"],
    -32013: [418, "I'm a teapot"],
    -32014: [501, "Not implemented"],
    -32015: [502, "Bad Gateway"],
    -32016: [503, "Service Unavailable"],
    -32017: [504, "Gateway Timeout"],
    -32018: [401, "Queue Limit Reached"]
}


class WappstoObjectType(str, Enum):
    """The different wappsto Object type names, used in a JSONRpc."""

    NETWORK = "network"
    DEVICE = "device"
    VALUE = "value"
    STATE = "state"


ObjectType2BaseModel: Dict[WappstoObjectType, Union[
    Type[Network], Type[Device], Type[ValueUnion], Type[State], Type[LogValue]]
] = {
    WappstoObjectType.NETWORK: Network,
    WappstoObjectType.DEVICE: Device,
    WappstoObjectType.VALUE: ValueUnion,
    WappstoObjectType.STATE: Union[State, LogValue],
}


def url_parser(url: str) -> List[Tuple[WappstoObjectType, Optional[uuid.UUID]]]:
    """Parse the Wappsto Url, for wappsto Type & given UUID."""
    r_list: List[Tuple[WappstoObjectType, Optional[uuid.UUID]]] = []
    obj_type: Optional[WappstoObjectType] = None
    if url is None:
        raise ValueError("Url need to be Set.")
    parsed_url = url.split("?")[0]
    if parsed_url.startswith("/services/2.0/"):
        parsed_url = parsed_url.replace("/services/2.0", "")
    for self_type, self_id in pair_wise(parsed_url.split("/")[1:]):
        obj_type = WappstoObjectType(self_type)
        if self_id:
            r_list.append((obj_type, uuid.UUID(self_id)))
        else:
            r_list.append((obj_type, None))
            break
    return r_list


class WappstoMethod(str, Enum):
    """The different Wappsto methods allowed for the JSONRpc."""

    GET = "GET"
    POST = "POST"
    PATCH = "PATCH"
    PUT = "PUT"
    DELETE = "DELETE"
    HEAD = "HEAD"


class Success(BaseModel):
    """The Default reply on a received JSONRpc."""

    success: bool = True

    model_config: ConfigDict = ConfigDict(extra='forbid')  # type: ignore


class Identifier(BaseModel):
    """The meta data structure for sending data."""

    identifier: Optional[str] = None  # UNSURE: Should this always be there?
    fast: Optional[bool] = None  # Default: False


class JsonMeta(BaseModel):
    """The meta data structure on received data."""

    server_send_time: datetime.datetime


class JsonReply(BaseModel):
    """The JSONRpc param structure for receiving data."""

    value: Optional[Union[
        Device,
        Network,
        State,
        ValueUnion,
        IdList,
        DeleteList,
        bool
    ]]
    meta: JsonMeta

    model_config: ConfigDict = ConfigDict(extra='forbid')  # type: ignore


class JsonData(BaseModel):
    """The JSONRpc param structure for sending."""

    url: str
    data: Optional[Any]
    # data: Optional[Union[
    #     Device,
    #     State,
    #     Network,
    #     ValueUnion,
    #     IdList,
    #     DeleteList,
    # ]]
    meta: Optional[Identifier]

    model_config: ConfigDict = ConfigDict(extra='forbid')  # type: ignore

    @field_validator('url')
    def path_check(cls, v: str) -> str:
        """Ensure that the url is valid."""
        url_parser(v)
        return v

    @field_validator("data", mode='before')
    def url_data_mapper(
        cls, v: Optional[Any], info: FieldValidationInfo
    ) -> Optional[
        Union[
            Network,
            Device,
            ValueUnion,
            State,
            LogValue,
            # IdList,
            # DeleteList,
        ]
    ]:
        """Check & enforce the data schema, depended on the method value."""
        if v is None:
            return v
        if type(v) in ObjectType2BaseModel.values():
            return v

        # TODO: handle IdList & DeleteList

        url_obj = url_parser(info.data['url'])
        obj_type = url_obj[-1][0]

        model = ObjectType2BaseModel[obj_type]
        if model is None:
            raise ValueError('Unhandled Object type.')

        if isinstance(model, BaseModel):
            return model.model_validate(v)

        model_converter: TypeAdapter = TypeAdapter(model)  # type: ignore
        return model_converter.validate_python(v)



"""Contain the Network Object."""
import logging
import uuid

from typing import Any
from typing import Dict
from typing import Callable
from typing import Optional

from ..service.template import ServiceClass
# from .service.rest_api import RestAPI

from .device import Device

from ..schema import base_schema as WSchema
from ..schema.iot_schema import WappstoMethod

from ..utils import name_check


# #############################################################################
#                                 Network Setup
# #############################################################################

class Network(object):
    """
    The root structure/grouping for the IoT Device.

    A network administrates the relationship between different devices.
    For example a network of lights.
    """

    schema = WSchema.Network

    def __init__(
        self,
        name: str,
        connection: ServiceClass,
        network_uuid: uuid.UUID,
        description: str = "",
    ) -> None:
        """Configure the Network settings."""
        self.log = logging.getLogger(__name__)
        self.log.addHandler(logging.NullHandler())

        kwargs = locals()
        self.__uuid: uuid.UUID = network_uuid
        self.element: WSchema.Network

        self.__callbacks: Dict[
            str,
            Callable[[WSchema.Network, WappstoMethod], None],
        ] = {}

        self.children_uuid_mapping: Dict[uuid.UUID, Device] = {}
        self.children_name_mapping: Dict[str, uuid.UUID] = {}

        # self.cloud_id_mapping: Dict[int, uuid.UUID] = {}

        self.connection: ServiceClass = connection

        self.element = self.schema(
            name=name,
            description=description,
            meta=WSchema.NetworkMeta(
                version=WSchema.WappstoVersion.V2_0,
                type=WSchema.NetworkMeta.WappstoMetaType.NETWORK,
                id=self.uuid
            )
        )

        element = self.connection.get_network(self.uuid)
        if element:
            self.__update_self(element)
            # self.log.debug(
            #     type(self.element.meta)
            # )
            # self.log.debug(
            #     self.element.meta
            # )
            # self.log.debug(
            #     type(element.meta)
            # )
            # self.log.debug(
            #     element.meta
            # )
            if self.element != element:
                # TODO: Post diff only.
                self.log.info("Data Models Differ. Sending Local.")
                self.connection.post_network(self.element)
        else:
            self.connection.post_network(self.element)

    @property
    def name(self) -> Optional[str]:
        """Returns the name of the value."""
        return self.element.name

    @property
    def uuid(self) -> uuid.UUID:
        """Returns the name of the value."""
        return self.__uuid

    def __argumentCountCheck(self, callback: Callable[[Any], Any], requiredArgumentCount: int) -> bool:
        """Check if the required Argument count for given function fits."""
        allArgument: int = callback.__code__.co_argcount
        the_default_count: int = len(callback.__defaults__) if callback.__defaults__ is not None else 1
        mandatoryArguments: int = callback.__code__.co_argcount - the_default_count
        return (
            requiredArgumentCount <= allArgument and requiredArgumentCount >= mandatoryArguments
        )

    # -------------------------------------------------------------------------
    #   Save/Load helper methods
    # -------------------------------------------------------------------------

    def __update_self(self, element: WSchema.Network) -> None:
        # TODO(MBK): Check if new devices was added! & Check diff.
        # NOTE: If there was a diff, post local one.
        self.element = element.model_copy(update=self.element.model_dump(exclude_none=True))
        self.element.meta = element.meta.model_copy(update=self.element.meta)
        # for nr, device in enumerate(self.element.device):
        #     self.cloud_id_mapping[nr] = device

    # -------------------------------------------------------------------------
    #   Network 'on-' methods
    # -------------------------------------------------------------------------

    def onChange(
        self,
        callback: Callable[['Network'], None],
    ) -> Callable[['Network'], None]:
        """
        Configure a callback for when a change to the Network have occurred.

        # UNSURE(MBK): How should it get the data back?
        """
        if not self.__argumentCountCheck(callback, 1):
            raise TypeError("The onChange callback, is called with 1 argument.")

        def _cb(obj: WSchema.Network, method: WappstoMethod) -> None:
            try:
                if method == WappstoMethod.PUT:
                    callback(self)
            except Exception:
                self.log.exception("OnChange callback error.")
                raise

        self.__callbacks['onChange'] = _cb

        self.connection.subscribe_network_event(
            uuid=self.uuid,
            callback=_cb
        )

        return callback

    def cancelOnChange(self) -> None:
        """Cancel the callback set in onChange-method."""
        self.connection.unsubscribe_network_event(
            uuid=self.uuid,
            callback=self.__callbacks['onChange']
        )

    def onCreate(
        self,
        callback: Callable[['Network'], None],
    ) -> Callable[['Network'], None]:
        """Configure a callback for when a create have been make for the Device."""
        if not self.__argumentCountCheck(callback, 1):
            raise TypeError("The onCreate callback, is called with 1 argument.")

        def _cb(obj: WSchema.Network, method: WappstoMethod) -> None:
            try:
                if method == WappstoMethod.POST:
                    callback(self)
            except Exception:
                self.log.exception("onCreate callback error.")
                raise

        self.__callbacks['onCreate'] = _cb

        self.connection.subscribe_network_event(
            uuid=self.uuid,
            callback=_cb
        )

        return callback

    def cancelOnCreate(self) -> None:
        """Cancel the callback set in onCreate-method."""
        self.connection.unsubscribe_network_event(
            uuid=self.uuid,
            callback=self.__callbacks['onCreate']
        )

    def onRefresh(
        self,
        callback: Callable[['Network'], None],
    ) -> Callable[['Network'], None]:
        """
        Configure an action when a refresh Network have been Requested.

        Normally when a refresh have been requested on a Network, ...
        ...
        # It can not! there is no '{"status":"update"}' that can be set.
        """
        if not self.__argumentCountCheck(callback, 1):
            raise TypeError("The onRefresh callback, is called with 1 argument.")

        def _cb(obj: WSchema.Network, method: WappstoMethod) -> None:
            try:
                if method == WappstoMethod.GET:
                    callback(self)
            except Exception:
                self.log.exception("onRefresh callback error.")
                raise

        self.__callbacks['onRefresh'] = _cb

        self.connection.subscribe_network_event(
            uuid=self.uuid,
            callback=_cb
        )

        return callback

    def cancelOnRefresh(self) -> None:
        """Cancel the callback set in onRefresh-method."""
        self.connection.unsubscribe_network_event(
            uuid=self.uuid,
            callback=self.__callbacks['onRefresh']
        )

    def onDelete(
        self,
        callback: Callable[['Network'], None],
    ) -> Callable[['Network'], None]:
        """
        Configure an action when a Delete Network have been Requested.

        Normally when a Delete have been requested on a Network,
        it is when it is not wanted anymore, and the Network have been
        unclaimed. Which mean that all the devices & value have to be
        recreated, and/or the program have to close.
        """
        if not self.__argumentCountCheck(callback, 1):
            raise TypeError("The onDelete callback, is called with 1 argument.")

        def _cb(obj: WSchema.Network, method: WappstoMethod) -> None:
            try:
                if method == WappstoMethod.DELETE:
                    callback(self)
            except Exception:
                self.log.exception("onDelete callback error.")
                raise

        self.__callbacks['onDelete'] = _cb

        self.connection.subscribe_network_event(
            uuid=self.uuid,
            callback=_cb
        )

        return callback

    def cancelOnDelete(self) -> None:
        """Cancel the callback set in onDelete-method."""
        self.connection.unsubscribe_network_event(
            uuid=self.uuid,
            callback=self.__callbacks['onDelete']
        )

    # -------------------------------------------------------------------------
    #   Network methods
    # -------------------------------------------------------------------------

    def refresh(self) -> None:
        """Not implemented."""
        raise NotImplementedError("Method: 'refresh' is not Implemented.")

    def change(self) -> None:
        """Not implemented."""
        pass

    def delete(self) -> None:
        """
        Normally it is used to unclaim a Network & delete all children.

        If a network delete itself, it will prompt a factory reset.
        This means that manufacturer and owner will be reset (or not),
        in relation of the rules set up in the certificates.
        """
        self.connection.delete_network(uuid=self.uuid)

    # -------------------------------------------------------------------------
    #   Create methods
    # -------------------------------------------------------------------------

    def createDevice(
        self,
        name: str,
        manufacturer: Optional[str] = None,
        product: Optional[str] = None,
        version: Optional[str] = None,
        serial: Optional[str] = None,
        protocol: Optional[str] = None,
        communication: Optional[str] = None,
        description: Optional[str] = None,
    ) -> Device:
        """
        Create a new Wappsto Device.

        A Wappsto Device is references something that is attached to the network
        that can be controlled or have values that can be reported to Wappsto.

        This could be a button that is connected to this unit,
        or in the case of this unit is a gateway, it could be a remote unit.
        """
        kwargs = locals()
        kwargs.pop('self')

        illegal_chars: str = name_check.illegal_characters(name)

        if illegal_chars:
            raise ValueError(
                f"Given name contain a illegal character: {illegal_chars}\n"
                f"May only contain: {name_check.wappsto_letters}"
            )

        device_uuid = self.connection.get_device_where(
            network_uuid=self.uuid,
            name=name
        )

        device_obj = Device(
            parent=self,
            device_uuid=device_uuid,
            **kwargs
        )
        self.__add_device(device_obj, kwargs['name'])
        return device_obj

    def __add_device(self, device: Device, name: str) -> None:
        """Helper function for Create, to only localy create it."""
        self.children_uuid_mapping[device.uuid] = device
        self.children_name_mapping[name] = device.uuid

    def close(self) -> None:
        """Do nothing, only here for compatibility."""
        pass



"""Contain the Socket ABC classes."""
from abc import ABC
from abc import abstractmethod

from threading import Lock
from enum import Enum

from typing import Union
from typing import Any
from typing import Callable
from typing import Optional


class StatusID(str, Enum):
    """The difference connection Statuses."""
    CONNECTING = "Connecting"
    CONNECTED = "Connected"
    DISCONNECTING = "Disconnecting"
    DISCONNETCED = "Disconnected"


class Connection(ABC):
    """."""

    send_ready: Lock

    @abstractmethod
    def send(
        self,
        data: Union[str, bytes]
    ) -> bool:
        """
        Send the str/Bytes to the server.

        If given string, it is encoded as 'uft-8' & send.

        Returns:
            True, if the data could be send else
            False.
        """

    @abstractmethod
    def receive(
        self,
        parser: Callable[[bytes], Any],
    ) -> Any:
        """
        Socket receive method.

        Method that handles receiving data from a socket. Capable of handling
        data chunks.

        Args:
            Callable: A parser, that returns the parsed data.
                      On Parsing Error, it should raise a
                      ValueError TypeError or any subClasses of those.
                      (Like 'JSONDecodeError' & 'pydantic.ValidationError' is)

        Returns:
            The Parsers output.
        """
        pass

    @abstractmethod
    def connect(self) -> Optional[bool]:
        """
        Connect to the server.

        Attempts a connection to the server on the provided address and port.

        Returns:
            'True' if the connection was successful else
            'False'
        """
        pass

    @abstractmethod
    def reconnect(
        self,
        retry_limit: Optional[int] = None
    ) -> bool:
        """
        Attempt to reconnect.

        Close the current connection, and then try to reconnect to the server,
        until the given amount of attempts, are above the retry_limit.
        If the retry_limit are not set, it will continue end.

        Args:
            retry_limit: the amount of retries, before it stops.

        Returns:
            'True' if the connection was successful else
            'False'
        """
        pass

    @abstractmethod
    def disconnect(self) -> None:
        """Disconnect from the server."""
        pass

    @abstractmethod
    def close(self) -> None:
        """
        Close the connection.

        Closes the socket object connection.
        """
        pass



"""Contain the Service class that handle the data conversion."""
from uuid import UUID
from enum import Enum

from typing import Callable
from typing import List
from typing import Optional
from typing import Union

from abc import ABC
from abc import abstractmethod

from ..schema.base_schema import BlobValue
from ..schema.base_schema import Device
from ..schema.base_schema import LogValue
from ..schema.base_schema import Network
from ..schema.base_schema import NumberValue
from ..schema.base_schema import State
from ..schema.base_schema import StringValue
from ..schema.base_schema import XmlValue
from ..schema.iot_schema import WappstoMethod


class StatusID(str, Enum):
    """The different states the service class can be in."""
    IDLE = "idle"
    # BATCHING = "batching"
    SENDING = "sending"
    SENDERROR = "send Error"
    SEND = "send"
    ERROR = "Error Reply"


class ServiceClass(ABC):
    """Should handle the conversion of the data from the socket and UI."""

    def close(self) -> None:
        """Closes the Service class down."""
        pass
    # #########################################################################
    #                               Helper API
    # #########################################################################

    def _resend_data(self, data: Union[str, bytes]) -> None:
        """For internal resending of data."""
        pass

    def ping(self) -> None:
        """Should send a ping to check the connection."""
        pass

    # #########################################################################
    #                               Network API
    # #########################################################################

    @abstractmethod
    def subscribe_network_event(
        self,
        uuid: UUID,
        callback: Callable[[Network, WappstoMethod], None]
    ) -> None:
        """Subscribe a function to be call on Network changes."""
        pass

    @abstractmethod
    def unsubscribe_network_event(
        self,
        uuid: UUID,
        callback: Callable[[Network, WappstoMethod], None]
    ) -> None:
        """Unsubscribe a function from Network changes."""
        pass

    @abstractmethod
    def post_network(self, data: Network) -> bool:
        """Create the network."""
        pass

    @abstractmethod
    def put_network(self, uuid: UUID, data: Network) -> bool:
        """Make changes to a network."""
        pass

    @abstractmethod
    def get_network(self, uuid: UUID) -> Union[Network, None]:
        """Request the network data."""
        pass

    @abstractmethod
    def delete_network(self, uuid: UUID) -> bool:
        """Remove the network."""
        pass

    # #########################################################################
    #                                Device API
    # #########################################################################

    @abstractmethod
    def subscribe_device_event(
        self,
        uuid: UUID,
        callback: Callable[[Device, WappstoMethod], None]
    ) -> None:
        """Subscribe a function to be call on given Device changes."""
        pass

    @abstractmethod
    def unsubscribe_device_event(
        self,
        uuid: UUID,
        callback: Callable[[Device, WappstoMethod], None]
    ) -> None:
        """Unsubscribe a function from given Device changes."""
        pass

    @abstractmethod
    def post_device(self, network_uuid: UUID, data: Device) -> bool:
        # url=f"/services/2.0/{uuid}/device",
        """Create given device."""
        pass

    @abstractmethod
    def put_device(self, uuid: UUID, data: Device) -> bool:
        """Make changes to a device."""
        pass

    @abstractmethod
    def get_device_where(self, network_uuid: UUID, **kwargs: str) -> Optional[UUID]:
        """Request data from a device with given values."""
        pass

    @abstractmethod
    def get_device(self, uuid: UUID) -> Union[Device, None]:
        """Request to get given device data."""
        pass

    @abstractmethod
    def delete_device(self, uuid: UUID) -> bool:
        """Remove to given device."""
        pass

    # #########################################################################
    #                                 Value API
    # #########################################################################

    ValueUnion = Union[StringValue, NumberValue, BlobValue, XmlValue]

    @abstractmethod
    def subscribe_value_event(
        self,
        uuid: UUID,
        callback: Callable[[ValueUnion, WappstoMethod], None]
    ) -> None:
        """Subscribe a function to be call on given value changes."""
        pass

    @abstractmethod
    def unsubscribe_value_event(
        self,
        uuid: UUID,
        callback: Callable[[ValueUnion, WappstoMethod], None]
    ) -> None:
        """Unsubscribe a function from given value changes."""
        pass

    @abstractmethod
    def post_value(self, device_uuid: UUID, data: ValueUnion) -> bool:
        """Create given value."""
        # url=f"/services/2.0/{uuid}/value",
        pass

    @abstractmethod
    def put_value(self, uuid: UUID, data: ValueUnion) -> bool:
        """Make changes to a value."""
        pass

    @abstractmethod
    def get_value_where(self, device_uuid: UUID, **kwargs: str) -> Optional[UUID]:
        """Request data from a value with given values."""
        pass

    @abstractmethod
    def get_value(self, uuid: UUID) -> Union[ValueUnion, None]:
        """Request to get given value data."""
        pass

    @abstractmethod
    def delete_value(self, uuid: UUID) -> bool:
        """Remove to given value."""
        pass

    # #########################################################################
    #                                State API
    # #########################################################################

    @abstractmethod
    def subscribe_state_event(
        self,
        uuid: UUID,
        callback: Callable[[State, WappstoMethod], None]
    ) -> None:
        """Subscribe a function to be call on given state changes."""
        pass

    @abstractmethod
    def unsubscribe_state_event(
        self,
        uuid: UUID,
        callback: Callable[[State, WappstoMethod], None]
    ) -> None:
        """Unsubscribe a function from given state changes."""
        pass

    @abstractmethod
    def post_state(self, value_uuid: UUID, data: Union[State, LogValue]) -> bool:
        """Create given state."""
        # url=f"/services/2.0/{uuid}/state",
        pass

    @abstractmethod
    def put_bulk_state(self, uuid: UUID, data: List[LogValue]) -> bool:
        """Make bulk changes the given state."""
        pass

    @abstractmethod
    def put_state(self, uuid: UUID, data: Union[State, LogValue]) -> bool:
        """Make changes to a state."""
        pass

    @abstractmethod
    def get_state(self, uuid: UUID) -> Union[State, None]:
        """Request to get given state data."""
        pass

    @abstractmethod
    def delete_state(self, uuid: UUID) -> bool:
        """Remove to given state."""
        pass



"""Contain the IoT Api that handle the data conversion."""
import copy
import json
import logging
import pathlib
import re
import threading

from uuid import UUID
from pathlib import Path

from concurrent.futures import ThreadPoolExecutor

from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import slxjsonrpc
from slxjsonrpc.schema.jsonrpc import ErrorModel

from .template import StatusID
from .template import ServiceClass

from ..schema.base_schema import Device
from ..schema.base_schema import IdList
from ..schema.base_schema import LogValue
from ..schema.base_schema import Network
from ..schema.base_schema import State
from ..schema.base_schema import ValueUnion
from ..schema.base_schema import WappstoObject

from ..schema.iot_schema import JsonData
from ..schema.iot_schema import Identifier
from ..schema.iot_schema import JsonReply
from ..schema.iot_schema import Success
from ..schema.iot_schema import WappstoMethod
# from ..schema.iot_schema import WappstoObjectType

from ..utils.certificateread import certificate_info_extraction
from ..utils import observer

from ..connections.sslsocket import TlsSocket
from ..connections.protocol import Connection


RpcSchemas = Union[
    slxjsonrpc.RpcBatch,
    slxjsonrpc.RpcError,
    slxjsonrpc.RpcNotification,
    slxjsonrpc.RpcRequest,
    slxjsonrpc.RpcResponse,
]

# POST   -> onCreate
# GET    -> onRefresh
# PUT    -> onChange
# DELETE -> onDelete

# create  -> POST
# refresh -> GET
# change  -> PUT
# delete  -> DELETE


class IoTAPI(ServiceClass):
    """IoTAPI, handle the conversion of the data from the socket and UI."""

    wappstoPort = {
        "dev": 52005,
        "qa": 53005,
        "staging": 54005,
        "prod": 443
    }

    def __init__(
        self,
        ca: Path,
        crt: Path,
        key: Path,
        worker_count: int = 2,
        fast_send: bool = False,
    ):
        """."""
        self.log = logging.getLogger(__name__)
        self.log.addHandler(logging.NullHandler())
        self.ca = ca
        self.crt = crt
        self.key = key

        self.timeout = 3

        self.fast_send = fast_send

        self.addr, self.port = self._url_gen(self.crt)

        self.connection: Connection

        self.connection = TlsSocket(
            address=self.addr,
            port=self.port,
            ca=self.ca,
            crt=self.crt,
            key=self.key
        )

        params = {
            x: JsonData for x in WappstoMethod
        }
        params[WappstoMethod.HEAD] = JsonData

        result = {
            WappstoMethod.GET: Union[JsonReply, Success],
            WappstoMethod.POST: Union[JsonReply, Success],
            WappstoMethod.PUT: Union[JsonReply, Success],
            WappstoMethod.DELETE: Union[JsonReply, Success],
            # WappstoMethod.PATCH: Union[JsonReply, Success],
            WappstoMethod.HEAD: Union[JsonReply, Success],
        }

        self.subscribers: Dict[
            UUID,
            List[Union[
                Callable[[Network, WappstoMethod], None],
                Callable[[Device, WappstoMethod], None],
                Callable[[ValueUnion, WappstoMethod], None],
                Callable[[State, WappstoMethod], None],
            ]]
        ] = {}

        method_cb = {
            WappstoMethod.GET: self._get,
            WappstoMethod.POST: self._post,
            # WappstoMethod.PATCH: self._patch,
            WappstoMethod.PUT: self._put,
            WappstoMethod.DELETE: self._delete,
            # WappstoMethod.HEAD: self._head,
        }

        self.connection.connect()

        self.jsonrpc = slxjsonrpc.SlxJsonRpc(
            methods=WappstoMethod,
            method_cb=method_cb,
            result=result,
            params=params,
        )
        self.jsonrpc._verbose = True

        self.killed = threading.Event()
        self.workers = ThreadPoolExecutor(max_workers=worker_count)

        self.workers.submit(self._receive_handler)

    def close(self) -> None:
        """Closes the IoTApi down."""
        self.killed.set()
        self.log.debug("Closing Connection.")
        self.connection.close()
        self.log.debug("Closing Workers")
        # while self.workers._work_queue.qsize() > 1:
        #     time.sleep(0.02)
        self.workers.shutdown()
        self.log.debug("IoTAPI Closed.")

    # #########################################################################
    #                              Helper Methods
    # #########################################################################

    def _url_gen(self, crt: pathlib.Path) -> Tuple[str, int]:
        cer = certificate_info_extraction(crt_path=crt)
        endpoint = cer.get('issuer', {}).get('commonName')
        port = self.wappstoPort.get(endpoint.split('.')[0], 443)
        if endpoint.split('.')[0] in self.wappstoPort.keys():
            addr = endpoint
        else:
            addr = f"collector.{endpoint}"

        if re.search(r'^[A-Za-z0-9+.-]+://', addr):
            addr = addr.split("://", maxsplit=1)[-1]
        # if not re.search(r':[0-9]+', addr):
        #     addr = f'{addr}:{port}'
        return (addr, port)

    def _receive_handler(self) -> None:
        self.log.debug("Receive Handler Started!")
        while not self.killed.is_set():
            try:
                data = self.connection.receive(parser=json.loads)

                if not data:
                    continue

                if isinstance(data, str):
                    self.log.warning(f'Received non JSONRPC data: {data}')
                    continue

                if not isinstance(data, list):
                    data = [data]
                for elemt in data:
                    _ids = [elemt.get('id', elemt) for elemt in data]
                    self.log.debug(f"Package received: {elemt.get('id', elemt)}")

                reply = self.jsonrpc.parser(data)
                self.log.debug(f"Package ID: {_ids}; Reply: {reply}")

                if not reply:
                    continue

                self._send_logic(reply)
                # UNSURE: How do we check if it was send?
                observer.post(StatusID.SEND, reply)

            except Exception:
                self.log.error(f"data: {data}")
                self.log.exception("Receive Handler Error:")
        self.log.debug("Receive Handler Stopped!")

    def _send_logic(self, data: RpcSchemas) -> None:
        # NOTE (MBK): Something do not work here!
        # if not data:
        #     return
        with self.connection.send_ready:  # NOTE: Waiting here, until ready!
            with self.jsonrpc.batch():
                batch_size = self.jsonrpc.batch_size()
                if batch_size:
                    send_data = self.jsonrpc.get_batch_data(data)
                    self.log.debug(f"Batching: {batch_size}")
                else:
                    send_data = copy.copy(data)

                if isinstance(send_data, slxjsonrpc.RpcBatch):
                    _id = "; ".join(x.id for x in send_data.root)
                elif send_data:
                    _id = send_data.id

                if _id:
                    self.log.debug(f"Package ID: {_id};")

                if send_data:
                    observer.post(StatusID.SENDING, send_data)
                    self.connection.send(send_data.model_dump_json(exclude_none=True))

    def _resend_data(self, data: Union[str, bytes]) -> None:
        j_data = json.loads(data)
        _cb_event = threading.Event()

        if not isinstance(j_data, list):
            j_data = [j_data]

        for l_data in j_data:
            rpc_id = l_data.get('id')

            if l_data.get('params'):
                s_data = self.jsonrpc.create_request(
                    callback=lambda data: _cb_event.set(),
                    error_callback=lambda err_data: _cb_event.set(),
                    method=l_data.get('method'),
                    params=JsonData(
                        data=l_data.get('params').get('data'),
                        url=l_data.get('params').get('url'),
                        meta=Identifier(fast=True, identifier=None)
                    ),
                )
                self._send_logic(s_data)
            elif l_data.get('result'):
                self.jsonrpc._id_cb[rpc_id] = lambda data: _cb_event.set()
                self.jsonrpc._id_error_cb[rpc_id] = lambda err_data: _cb_event.set()
                self.jsonrpc._id_method[rpc_id] = l_data.get('method')
                # self.jsonrpc._add_result_handling(  # NOTE: slxJsonRpc v0.8.1.
                #     method=l_data.get('method'),
                #     _id=rpc_id,
                #     callback=lambda data: _cb_event(),
                #     error_callback=lambda err_data: _cb_event(),
                # )
                s_data = self.jsonrpc._parse_data(l_data)
                self._send_logic(s_data)

            self.log.debug(f"--CALLBACK Ready! {rpc_id}")
            if _cb_event.wait(timeout=self.timeout):
                self.log.debug(f"--CALLBACK EVENT! {rpc_id}")
                observer.post(StatusID.SEND, s_data)
            else:
                self.log.debug(f"--CALLBACK None! {rpc_id}")
                observer.post(StatusID.SENDERROR, s_data)
        return None

    # -------------------------------------------------------------------------
    #                               API Helpers
    # -------------------------------------------------------------------------

    def __create_json_data(
        self,
        data: List[Any],
        url: str,
        method: WappstoMethod,
    ) -> RpcSchemas:
        _cb_event = threading.Event()
        _err_data: Optional[ErrorModel] = None

        def _err_callback(err_data: ErrorModel) -> None:
            nonlocal _err_data
            nonlocal _cb_event
            _err_data = err_data
            _cb_event.set()

        with self.jsonrpc.batch():
            for values in data:
                j_data = JsonData(
                    data=values,
                    url=url,
                    meta=Identifier(fast=True, identifier=None)
                    if self.fast_send and method != WappstoMethod.GET
                    else None
                )

                self.log.debug(f"Sending for: {url}")

                self.jsonrpc.create_request(
                    method=method,
                    callback=lambda data: _cb_event.set(),
                    error_callback=_err_callback,
                    params=j_data
                )

        rpc_data = self.jsonrpc.get_batch_data()

        return rpc_data, _cb_event, _err_data

    def _no_reply_bulk_send(
        self,
        data: List[Any],
        url: str,
        method: WappstoMethod,
    ) -> bool:
        rpc_data, _cb_event, _err_data = self.__create_json_data(
            data=data, url=url, method=method,
        )
        rpc_id = "[" + ",".join(
            [f'"{getattr(rpc_data, "id", None)}"' for rpc_d in rpc_data]
        ) + "]"

        self._send_logic(rpc_data)

        self.log.debug(f"--CALLBACK Ready! {rpc_id}")
        if _cb_event.wait(timeout=self.timeout):
            if _err_data:
                self.log.debug(f"--CALLBACK Error! {_err_data}")
                observer.post(StatusID.ERROR, _err_data)
                # raise ConnectionError(_err_data.message)
                return False
            self.log.debug(f"--CALLBACK EVENT! {rpc_id}")
            observer.post(StatusID.SEND, rpc_data)
            return True
        observer.post(StatusID.SENDERROR, rpc_data)
        self.log.debug(f"--CALLBACK None! {rpc_id}")
        raise TimeoutError(f'JsonRPC reply timeout on package: {rpc_id}')

    def _no_reply_send(
        self,
        data: Optional[Union[WappstoObject, LogValue]],
        url: str,
        method: WappstoMethod,
    ) -> bool:
        j_data = JsonData(
            data=data,
            url=url,
            meta=Identifier(fast=True, identifier=None)
            if self.fast_send and method != WappstoMethod.GET else None
        )

        self.log.debug(f"Sending for: {url}")

        _cb_event = threading.Event()
        _err_data: Optional[ErrorModel] = None

        def _err_callback(err_data: ErrorModel) -> None:
            nonlocal _err_data
            nonlocal _cb_event
            _err_data = err_data
            _cb_event.set()

        rpc_data = self.jsonrpc.create_request(
            method=method,
            callback=lambda data: _cb_event.set(),
            error_callback=_err_callback,
            params=j_data
        )

        rpc_id = getattr(rpc_data, 'id', None)

        self._send_logic(rpc_data)

        self.log.debug(f"--CALLBACK Ready! {rpc_id}")
        if _cb_event.wait(timeout=self.timeout):
            if _err_data:
                self.log.debug(f"--CALLBACK Error! {_err_data}")
                observer.post(StatusID.ERROR, _err_data)
                # raise ConnectionError(_err_data.message)
                return False
            self.log.debug(f"--CALLBACK EVENT! {rpc_id}")
            observer.post(StatusID.SEND, rpc_data)
            return True
        observer.post(StatusID.SENDERROR, rpc_data)
        self.log.debug(f"--CALLBACK None! {rpc_id}")
        raise TimeoutError(f'JsonRPC reply timeout on package: {rpc_id}')

    def _reply_send(
        self,
        data: Optional[WappstoObject],
        url: str,
        method: WappstoMethod,
    ) -> Optional[WappstoObject]:
        j_data = JsonData(
            data=data,
            url=url,
            meta=Identifier(fast=True, identifier=None)
            if self.fast_send and method != WappstoMethod.GET else None
        )

        self.log.debug(f"Sending for: {url}")

        _cb_event = threading.Event()
        _err_data: ErrorModel = None

        def _err_callback(err_data: ErrorModel) -> None:
            nonlocal _err_data
            nonlocal _cb_event
            _err_data = err_data
            _cb_event.set()

        _data: Optional[JsonReply] = None

        def _data_callback(data: JsonReply) -> None:
            nonlocal _data
            nonlocal _cb_event
            _data = data
            _cb_event.set()

        rpc_data = self.jsonrpc.create_request(
            method=method,
            callback=_data_callback,
            error_callback=_err_callback,
            params=j_data
        )

        rpc_id = getattr(rpc_data, 'id', None)

        self._send_logic(rpc_data)

        self.log.debug(f"--CALLBACK Ready! {rpc_id}")
        if _cb_event.wait(timeout=self.timeout):
            if _data:
                self.log.debug(f"--CALLBACK EVENT! {rpc_id}")
                observer.post(StatusID.SEND, rpc_data)
                return _data.value
            if _err_data:
                self.log.warning(f"--CALLBACK Error! {_err_data}")
                observer.post(StatusID.ERROR, _err_data)
                # raise ConnectionError(_err_data.message)
                return None
        self.log.debug(f"--CALLBACK None! {rpc_id}")
        observer.post(StatusID.SENDERROR, rpc_data)
        raise TimeoutError(f'JsonRPC reply timeout on package: {rpc_id}')

    # -------------------------------------------------------------------------
    #                              Callback Helpers
    # -------------------------------------------------------------------------

    def _cb_handler(self, data: JsonData, method: WappstoMethod) -> None:
        object_uuid = UUID(data.url.split("/")[-1])
        self.log.debug(f"Object UUID: {object_uuid}")
        for cb in self.subscribers.get(object_uuid, [self._default_cb]):
            self.workers.submit(cb, data.data, method)
            self.log.debug(f"Submitted to Worker: {cb}")

    def _default_cb(self, data: WappstoObject, method: WappstoMethod) -> None:
        self.log.warning(
            f"No callback found for method: {method}; data {data}"
        )

    # #########################################################################
    #                              Callback Methods
    # #########################################################################

    def _get(self, data: JsonData) -> Union[Success, JsonData]:
        # Add data and callback-function to a threadpool.
        self.log.debug("_get: Called!")
        self._cb_handler(data=data, method=WappstoMethod.GET)
        return Success()

    def _post(self, data: JsonData) -> Union[Success, JsonData]:
        # Add data and callback-function to a threadpool.
        self.log.debug("_post: Called!")
        self._cb_handler(data=data, method=WappstoMethod.POST)
        return Success()

    def _put(self, data: JsonData) -> Union[Success, JsonData]:
        # Add data and callback-function to a threadpool.
        self.log.debug("_put: Called!")
        self._cb_handler(data=data, method=WappstoMethod.PUT)
        return Success()

    def _delete(self, data: JsonData) -> Union[Success, JsonData]:
        # Add data and callback-function to a threadpool.
        self.log.debug("_delete: Called!")
        self._cb_handler(data=data, method=WappstoMethod.DELETE)
        return Success()

    # def _patch(self, data: JsonData) -> Union[Success, JsonData]:
    #     # Add data and callback-function to a threadpool.
    #     self.log.debug("_patch: Called!")
    #     self._cb_handler(data=data, method=WappstoMethod.PATCH)
    #     return Success()

    # def _head(self, data: JsonData) -> Union[Success, JsonData]:
    #     # Add data and callback-function to a threadpool.
    #     self.log.debug("_head: Called!")
    #     self._cb_handler(data=data, method=WappstoMethod.HEAD)
    #     return Success()

    # #########################################################################
    #                               Helper API
    # #########################################################################

    def ping(self) -> None:
        """Sends a ping to check the connection."""
        return self._no_reply_send(
            data=None,
            url="/network",
            method=WappstoMethod.HEAD
        )

    # #########################################################################
    #                               Network API
    # #########################################################################

    def subscribe_network_event(
        self,
        uuid: UUID,
        callback: Callable[[Network, WappstoMethod], None]
    ) -> None:
        """Subscribe a function to be call on Network changes."""
        self.subscribers.setdefault(uuid, []).append(callback)

    def unsubscribe_network_event(
        self,
        uuid: UUID,
        callback: Callable[[Network, WappstoMethod], None]
    ) -> None:
        """Unsubscribe a function from Network changes."""
        self.subscribers.get(uuid, []).remove(callback)

    def post_network(self, data: Network) -> bool:
        """Create the network."""
        return self._no_reply_send(
            data=data,
            url="/network/",
            method=WappstoMethod.POST
        )

    def put_network(self, uuid: UUID, data: Network) -> bool:
        """Make changes to a network."""
        return self._no_reply_send(
            data=data,
            url=f"/network/{uuid}",
            method=WappstoMethod.PUT
        )

    def get_network(self, uuid: UUID) -> Optional[Network]:
        """Request the network data."""
        return self._reply_send(
            data=None,  # NOTE: Should be nonexistent or Null.
            url=f"/network/{uuid}",
            method=WappstoMethod.GET
        )

    def delete_network(self, uuid: UUID) -> bool:
        """Remove the network."""
        return self._no_reply_send(
            data=None,  # NOTE: Should be nonexistent or Null.
            url=f"/network/{uuid}",
            method=WappstoMethod.DELETE
        )

    # #########################################################################
    #                                Device API
    # #########################################################################

    def subscribe_device_event(
        self,
        uuid: UUID,
        callback: Callable[[Device, WappstoMethod], None]
    ) -> None:
        """Subscribe a function to be call on given Device changes."""
        self.subscribers.setdefault(uuid, []).append(callback)

    def unsubscribe_device_event(
        self,
        uuid: UUID,
        callback: Callable[[Device, WappstoMethod], None]
    ) -> None:
        """Unsubscribe a function from given Device changes."""
        self.subscribers.get(uuid, []).remove(callback)

    def post_device(self, network_uuid: UUID, data: Device) -> bool:
        """Create given device."""
        return self._no_reply_send(
            data=data,
            url=f"/network/{network_uuid}/device/",
            method=WappstoMethod.POST
        )

    def put_device(self, uuid: UUID, data: Device) -> bool:
        """Make changes to a device."""
        return self._no_reply_send(
            data=data,
            url=f"/device/{uuid}",
            method=WappstoMethod.PUT
        )

    def get_device_where(self, network_uuid: UUID, **kwargs: str) -> Optional[UUID]:
        """Request data from a device with given values."""
        # /network/{uuid}/device?this_name==X
        key, value = list(kwargs.items())[0]
        url = f"/network/{network_uuid}/device?this_{key}=={value}"
        data: IdList = self._reply_send(
            data=None,
            url=url,
            method=WappstoMethod.GET
        )

        temp = getattr(data, "id", None)
        if not temp:
            return None
        return temp[0]

    def get_device(self, uuid: UUID) -> Union[Device, None]:
        """Request to get given device data."""
        return self._reply_send(
            data=None,
            url=f"/device/{uuid}",
            method=WappstoMethod.GET
        )

    def delete_device(self, uuid: UUID) -> bool:
        """Remove to given device."""
        return self._no_reply_send(
            data=None,
            url=f"/device/{uuid}",
            method=WappstoMethod.DELETE
        )

    # #########################################################################
    #                                 Value API
    # #########################################################################

    def subscribe_value_event(
        self,
        uuid: UUID,
        callback: Callable[[ValueUnion, WappstoMethod], None]
    ) -> None:
        """Subscribe a function to be call on given value changes."""
        self.subscribers.setdefault(uuid, []).append(callback)

    def unsubscribe_value_event(
        self,
        uuid: UUID,
        callback: Callable[[ValueUnion, WappstoMethod], None]
    ) -> None:
        """Unsubscribe a function from given value changes."""
        self.subscribers.get(uuid, []).remove(callback)

    def post_value(self, device_uuid: UUID, data: ValueUnion) -> bool:
        """Create given value."""
        # url=f"/services/2.0/{uuid}/value",
        return self._no_reply_send(
            data=data,
            url=f"/device/{device_uuid}/value/",
            method=WappstoMethod.POST
        )

    def put_value(self, uuid: UUID, data: ValueUnion) -> bool:
        """Make changes to a value."""
        # url=f"/services/2.0/value/{uuid}",
        return self._no_reply_send(
            data=data,
            url=f"/value/{uuid}",
            method=WappstoMethod.PUT
        )

    def get_value_where(self, device_uuid: UUID, **kwargs: str) -> Optional[UUID]:
        """Request data from a value with given values."""
        # /network/{uuid}/device?this_name==X
        key, value = list(kwargs.items())[0]
        url = f"/device/{device_uuid}/value?this_{key}=={value}"
        data: IdList = self._reply_send(
            data=None,
            url=url,
            method=WappstoMethod.GET
        )

        temp: Optional[List[UUID]] = getattr(data, "id", None)
        if not temp:
            return None
        return temp[0]

    def get_value(self, uuid: UUID) -> Union[ValueUnion, None]:
        """Request to get given value data."""
        # url=f"/services/2.0/value/{uuid}",
        return self._reply_send(
            data=None,
            url=f"/value/{uuid}",
            method=WappstoMethod.GET
        )

    def delete_value(self, uuid: UUID) -> bool:
        """Remove to given value."""
        # url=f"/services/2.0/value/{uuid}",
        return self._no_reply_send(
            data=None,
            url=f"/value/{uuid}",
            method=WappstoMethod.DELETE
        )

    # #########################################################################
    #                                State API
    # #########################################################################

    def subscribe_state_event(
        self,
        uuid: UUID,
        callback: Callable[[State, WappstoMethod], None]
    ) -> None:
        """Subscribe a function to be call on given state changes."""
        self.subscribers.setdefault(uuid, []).append(callback)

    def unsubscribe_state_event(
        self,
        uuid: UUID,
        callback: Callable[[State, WappstoMethod], None]
    ) -> None:
        """Unsubscribe a function from given state changes."""
        self.subscribers.get(uuid, []).remove(callback)

    def post_state(self, value_uuid: UUID, data: Union[State, LogValue]) -> bool:
        """Create given state."""
        # url=f"/services/2.0/{uuid}/state",
        return self._no_reply_send(
            data=data,
            url=f"/value/{value_uuid}/state/",
            method=WappstoMethod.POST
        )

    def put_bulk_state(self, uuid: UUID, data: List[LogValue]) -> bool:
        """Make bulk changes the given state."""
        # url=f"/services/2.0/state/{uuid}",
        try:
            return self._no_reply_bulk_send(
                data=data,
                url=f"/state/{uuid}",
                method=WappstoMethod.PUT
            )
        except TimeoutError:
            msg = (
                'Error in sending Values.\n'
                'If Offline storage is enabled it will be send later.'
            )
            self.log.exception(msg)

    def put_state(self, uuid: UUID, data: Union[State, LogValue]) -> bool:
        """Make changes to a state."""
        # url=f"/services/2.0/state/{uuid}",
        try:
            return self._no_reply_send(
                data=data,
                url=f"/state/{uuid}",
                method=WappstoMethod.PUT
            )
        except TimeoutError:
            msg = (
                'Error in sending Values.\n'
                'If Offline storage is enabled it will be send later.'
            )
            self.log.exception(msg)

    def get_state(self, uuid: UUID) -> Union[State, None]:
        """Request to get given state data."""
        return self._reply_send(
            data=None,
            url=f"/state/{uuid}",
            method=WappstoMethod.GET
        )

    def delete_state(self, uuid: UUID) -> bool:
        """Remove to given state."""
        # url=f"/services/2.0/state/{uuid}",
        return self._no_reply_send(
            data=None,
            url=f"/state/{uuid}",
            method=WappstoMethod.DELETE
        )



"""Contain the wappsto timestamp converters."""
import datetime


def timestamp() -> str:
    """
    Return The default timestamp used for Wappsto.

    The timestamp are always set to the UTC timezone.

    Returns:
        The UTC time string in ISO format.
    """
    return datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%fZ')


def str_to_datetime(timestamp: str) -> datetime.datetime:
    """
    Convert the logger timestamp to a ISO-T-format w/ timezone.

    Args:
        data_string: The timestamp needed to be converted.

    Returns:
        The converted timestamp.
    """
    return datetime.datetime.strptime(
        timestamp,
        '%Y-%m-%dT%H:%M:%S.%fZ'
    )



"""Contain Certificate helper functions."""
import ssl
import pathlib

from typing import Any
from typing import Dict
from typing import Union


def certificate_info_extraction(crt_path: Union[str, pathlib.Path]) -> Dict[str, Any]:
    """Parse the given certificate into a dictionary."""
    crt: dict = ssl._ssl._test_decode_cert(crt_path)
    crt['issuer'] = {x[0][0]: x[0][1] for x in crt['issuer']}
    crt['subject'] = {x[0][0]: x[0][1] for x in crt['subject']}
    return crt



"""Contain the Device Object."""
import uuid
import logging

from enum import Enum

from typing import Any
from typing import Callable
from typing import cast
from typing import Dict
from typing import Optional
from typing import Union

from ..service.template import ServiceClass

from ..schema import base_schema as WSchema
from ..schema.iot_schema import WappstoMethod
from ..schema.base_schema import PermissionType

from .value import Value
from .template import valueSettings
from .template import ValueTemplate
from .template import ValueBaseType

from ..utils import name_check

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    # NOTE: To avoid circle import
    from .network import Network


# #############################################################################
#                                 Device Setup
# #############################################################################

# class RequestType(str, Enum):
#     refresh = "refresh"
#     delete = "delete"


class ChangeType(str, Enum):
    """List of the different value that can be Changed."""
    value = "value"
    name = "name"
    manufacturer = "manufacturer"
    product = "product"
    version = "version"
    serial = "serial"
    description = "description"


class Device:
    """
    A Device is an hardware tool that own different Values.

    By using an example a network of light is structured
    with different light bulbs (devices).
    These devices own their own values of brightness, energy and color.
    """

    schema = WSchema.Device

    def __init__(
        self,
        parent: 'Network',
        device_uuid: Optional[uuid.UUID],
        name: Optional[str] = None,
        manufacturer: Optional[str] = None,
        product: Optional[str] = None,
        version: Optional[str] = None,
        protocol: Optional[str] = None,
        communication: Optional[str] = None,
        serial: Optional[str] = None,
        description: Optional[str] = None,
    ):
        """Configure the Device settings."""
        self.log: logging.Logger = logging.getLogger(__name__)
        self.log.addHandler(logging.NullHandler())

        self.__callbacks: Dict[str, Callable[[WSchema.Device, WappstoMethod], None]] = {}

        self.parent = parent
        self.element: WSchema.Device

        self.children_uuid_mapping: Dict[uuid.UUID, Value] = {}
        self.children_name_mapping: Dict[str, uuid.UUID] = {}

        # self.cloud_id_mapping: Dict[int, uuid.UUID] = {}

        self.connection: ServiceClass = parent.connection

        element = self.connection.get_device(device_uuid) if device_uuid else None

        self.__uuid: uuid.UUID = device_uuid if device_uuid else uuid.uuid4()

        self.element = self.schema(
            name=name,
            manufacturer=manufacturer,
            product=product,
            version=version,
            serial=serial,
            description=description,
            protocol=protocol,
            communication=communication,
            meta=WSchema.DeviceMeta(
                version=WSchema.WappstoVersion.V2_0,
                type=WSchema.DeviceMeta.WappstoMetaType.DEVICE,
                id=self.uuid
            )
        )

        if element:
            self.__update_self(element)
            # self.log.debug(
            #     self.element
            # )
            # self.log.debug(
            #     element
            # )
            if self.element != element:
                # TODO: Post diff only.
                self.log.info("Data Models Differ. Sending Local.")
                self.connection.post_device(
                    network_uuid=self.parent.uuid,
                    data=self.element,
                )
        else:
            self.connection.post_device(
                network_uuid=self.parent.uuid,
                data=self.element
            )

    @property
    def name(self) -> str:
        """Returns the name of the value."""
        return cast(str, self.element.name)

    @property
    def uuid(self) -> uuid.UUID:
        """Returns the name of the value."""
        return self.__uuid

    def __update_self(self, element: WSchema.Device) -> None:
        # TODO(MBK): Check if new devices was added! & Check diff.
        # NOTE: If there was a diff, post local one.
        self.element = element.model_copy(update=self.element.model_dump(exclude_none=True))
        self.element.meta = element.meta.model_copy(update=self.element.meta)
        # for nr, value in enumerate(self.element.value):
        #     self.cloud_id_mapping[nr] = value

    def __argumentCountCheck(self, callback: Callable[[Any], Any], requiredArgumentCount: int) -> bool:
        """Check if the required Argument count for given function fits."""
        allArgument: int = callback.__code__.co_argcount
        the_default_count: int = len(callback.__defaults__) if callback.__defaults__ is not None else 1
        mandatoryArguments: int = callback.__code__.co_argcount - the_default_count
        return (
            requiredArgumentCount <= allArgument and requiredArgumentCount >= mandatoryArguments
        )

    # -------------------------------------------------------------------------
    #   Device 'on-' methods
    # -------------------------------------------------------------------------

    def onDelete(
        self,
        callback: Callable[['Device'], None],
    ) -> Callable[['Device'], None]:
        """
        Configure an action when a Delete on this Device have been Requested.

        Normally when a Delete have been requested,
        it is when it is not wanted anymore.
        Which mean that all the device and it's values have to be removed,
        and the process of setting up the device, should be executed again.
        This could result in the same device are created again, or if the
        device was not found, it will just be removed.
        """
        if not self.__argumentCountCheck(callback, 1):
            raise TypeError("The onDelete callback, is called with 1 argument.")

        def _cb(obj: WSchema.Device, method: WappstoMethod) -> None:
            try:
                if method in WappstoMethod.DELETE:
                    callback(self)
            except Exception:
                self.log.exception("onDelete callback error.")
                raise

        self.__callbacks['onDelete'] = _cb

        self.connection.subscribe_device_event(
            uuid=self.uuid,
            callback=_cb
        )

        return callback

    def cancelOnDelete(self) -> None:
        """Cancel the callback set in onDelete-method."""
        self.connection.unsubscribe_device_event(
            uuid=self.uuid,
            callback=self.__callbacks['onDelete']
        )

    def onRefresh(
        self,
        callback: Callable[['Device'], None],
    ) -> Callable[['Device'], None]:
        """
        Add trigger for when a Refresh where requested.

        # It can not! there is no '{"status":"update"}' that can be set.

        Callback:
            ValueObj: This object that have had a refresh request for.
        """
        if not self.__argumentCountCheck(callback, 1):
            raise TypeError("The onRefresh callback, are called with 1 argument.")

        def _cb(obj: WSchema.Device, method: WappstoMethod) -> None:
            try:
                if method in WappstoMethod.GET:
                    callback(self)
            except Exception:
                self.log.exception("onRefresh callback error.")
                raise

        self.__callbacks['onRefresh'] = _cb

        self.connection.subscribe_device_event(
            uuid=self.uuid,
            callback=_cb
        )

        return callback

    def cancelOnRefresh(self) -> None:
        """Cancel the callback set in onRefresh-method."""
        self.connection.unsubscribe_device_event(
            uuid=self.uuid,
            callback=self.__callbacks['onRefresh']
        )

    def onChange(
        self,
        callback: Callable[['Device'], None],
    ) -> Callable[['Device'], None]:
        """Configure a callback for when a change to the Device have occurred."""
        if not self.__argumentCountCheck(callback, 1):
            raise TypeError("The onChange callback, are called with 1 argument.")

        def _cb(obj: WSchema.Device, method: WappstoMethod) -> None:
            try:
                if method in WappstoMethod.PUT:
                    callback(self)
            except Exception:
                self.log.exception("OnChange callback error.")
                raise

        self.__callbacks['onChange'] = _cb

        self.connection.subscribe_device_event(
            uuid=self.uuid,
            callback=_cb
        )

        return callback

    def cancelOnChange(self) -> None:
        """Cancel the callback set in onChange-method."""
        self.connection.unsubscribe_device_event(
            uuid=self.uuid,
            callback=self.__callbacks['onChange']
        )

    def onCreate(
        self,
        callback: Callable[['Device'], None],
    ) -> Callable[['Device'], None]:
        """Configure a callback for when a request have been make for the Value."""
        if not self.__argumentCountCheck(callback, 1):
            raise TypeError("The onCreate callback, are called with 1 argument.")

        def _cb(obj: WSchema.Device, method: WappstoMethod) -> None:
            try:
                if method in WappstoMethod.POST:
                    callback(self)
            except Exception:
                self.log.exception("onCreate callback error.")
                raise

        self.__callbacks['onCreate'] = _cb

        self.connection.subscribe_device_event(
            uuid=self.uuid,
            callback=_cb
        )

        return callback

    def cancelOnCreate(self) -> None:
        """Cancel the callback set in onCreate-method."""
        self.connection.unsubscribe_device_event(
            uuid=self.uuid,
            callback=self.__callbacks['onCreate']
        )

    # -------------------------------------------------------------------------
    #   Device methods
    # -------------------------------------------------------------------------

    def refresh(self) -> None:
        """Not Implemented."""
        raise NotImplementedError("Method: 'refresh' is not Implemented.")

    def change(self, change_type: ChangeType) -> None:
        """
        Update a parameter in the Device structure.

        A parameter that a device can have that can be updated could be:
         - manufacturer
         - product
         - version
         - serial
         - description
        """
        pass

    def delete(self) -> None:
        """Request a delete of the Device, & all it's Children."""
        self.connection.delete_device(uuid=self.uuid)

    # -------------------------------------------------------------------------
    #   Other methods
    # -------------------------------------------------------------------------

    def createNumberValue(
        self,
        name: str,
        *,
        permission: PermissionType,
        type: str,
        min: Union[int, float],
        max: Union[int, float],
        step: Union[int, float],
        unit: str,
        description: Optional[str] = None,
        si_conversion: Optional[str] = None,
        period: Optional[int] = None,  # in Sec
        delta: Optional[Union[int, float]] = None,
        mapping: Optional[Dict[str, str]] = None,
        meaningful_zero: Optional[bool] = None,
        ordered_mapping: Optional[bool] = None,
    ) -> Value:
        """
        Create a Wappsto Number Value.

        A Wappsto Value is where the changing data can be found & are handled.

        This require you to setup manually, that `createValue`
        with `value_template` setup for you.
        """
        kwargs = locals()
        kwargs.pop('self')

        illegal_chars: str = name_check.illegal_characters(name)

        if illegal_chars:
            raise ValueError(
                f"Given name contain a illegal character: {illegal_chars}\n"
                f"May only contain: {name_check.wappsto_letters}"
            )

        value_uuid = self.connection.get_value_where(
            device_uuid=self.uuid,
            name=name
        )

        value_obj = Value(
            parent=self,
            value_uuid=value_uuid,
            value_type=ValueBaseType.NUMBER,
            **kwargs
        )

        self.__add_value(value_obj, kwargs['name'])
        return value_obj

    def createStringValue(
        self,
        name: str,
        *,
        permission: PermissionType,
        type: str,
        max: Union[int, float],
        encoding: Optional[str] = None,
        description: Optional[str] = None,
        period: Optional[int] = None,  # in Sec
        delta: Optional[Union[int, float]] = None,
    ) -> Value:
        """
        Create a Wappsto String Value.

        A Wappsto Value is where the changing data can be found & are handled.

        This require you to setup manually, that `createValue`
        with `value_template` setup for you.
        """
        kwargs = locals()
        kwargs.pop('self')

        illegal_chars: str = name_check.illegal_characters(name)

        if illegal_chars:
            raise ValueError(
                f"Given name contain a illegal character: {illegal_chars}\n"
                f"May only contain: {name_check.wappsto_letters}"
            )

        value_uuid = self.connection.get_value_where(
            device_uuid=self.uuid,
            name=name
        )

        value_obj = Value(
            parent=self,
            value_uuid=value_uuid,
            value_type=ValueBaseType.STRING,
            **kwargs
        )

        self.__add_value(value_obj, kwargs['name'])
        return value_obj

    def createBlobValue(
        self,
        name: str,
        *,
        permission: PermissionType,
        type: str,
        max: Union[int, float],
        encoding: Optional[str] = None,
        description: Optional[str] = None,
        period: Optional[int] = None,  # in Sec
        delta: Optional[Union[int, float]] = None,
    ) -> Value:
        """
        Create a Wappsto BLOB Value.

        A Wappsto Value is where the changing data can be found & are handled.

        This require you to setup manually, that `createValue`
        with `value_template` setup for you.
        """
        kwargs = locals()
        kwargs.pop('self')

        illegal_chars: str = name_check.illegal_characters(name)

        if illegal_chars:
            raise ValueError(
                f"Given name contain a illegal character: {illegal_chars}\n"
                f"May only contain: {name_check.wappsto_letters}"
            )

        value_uuid = self.connection.get_value_where(
            device_uuid=self.uuid,
            name=name
        )

        value_obj = Value(
            parent=self,
            value_uuid=value_uuid,
            value_type=ValueBaseType.BLOB,
            **kwargs
        )

        self.__add_value(value_obj, kwargs['name'])
        return value_obj

    def createXmlValue(
        self,
        name: str,
        *,
        permission: PermissionType,
        type: str,
        xsd: Optional[str] = None,
        namespace: Optional[str] = None,
        description: Optional[str] = None,
        period: Optional[int] = None,  # in Sec
        delta: Optional[Union[int, float]] = None,
    ) -> Value:
        """
        Create a Wappsto XML Value.

        A Wappsto Value is where the changing data can be found & are handled.

        This require you to setup manually, that `createValue`
        with `value_template` setup for you.
        """
        kwargs = locals()
        kwargs.pop('self')

        illegal_chars: str = name_check.illegal_characters(name)

        if illegal_chars:
            raise ValueError(
                f"Given name contain a illegal character: {illegal_chars}\n"
                f"May only contain: {name_check.wappsto_letters}"
            )

        value_uuid = self.connection.get_value_where(
            device_uuid=self.uuid,
            name=name
        )

        value_obj = Value(
            parent=self,
            value_uuid=value_uuid,
            value_type=ValueBaseType.XML,
            **kwargs
        )

        self.__add_value(value_obj, kwargs['name'])
        return value_obj

    def createValue(
        self,
        name: str,
        permission: PermissionType,
        value_template: ValueTemplate,
        description: Optional[str] = None,
    ) -> Value:
        """
        Create a Wappsto Value.

        A Wappsto Value is where the changing data can be found & are handled.

        If a value_template have been set, that means that the parameters like:
        name, permission, min, max, step, encoding & unit have been set
        for you, to be the right settings for the given type. But you can
        still change it, if you choose sow.
        """
        illegal_chars: str = name_check.illegal_characters(name)

        if illegal_chars:
            raise ValueError(
                f"Given name contain a illegal character: {illegal_chars}\n"
                f"May only contain: {name_check.wappsto_letters}"
            )

        value_uuid = self.connection.get_value_where(
            device_uuid=self.uuid,
            name=name
        )

        value_obj = Value(
            parent=self,
            name=name,
            value_uuid=value_uuid,
            permission=permission,
            **valueSettings[value_template].model_dump()
        )

        self.__add_value(value_obj, name)
        return value_obj

    def __add_value(self, value: Value, name: str) -> None:
        """Helper function for Create, to only localy create it."""
        self.children_uuid_mapping[value.uuid] = value
        self.children_name_mapping[name] = value.uuid

    def close(self) -> None:
        """Do nothing, only here for compatibility."""
        pass



from setuptools import find_packages
from setuptools import setup

import codecs
import os
import re


def get_long_description():
    long_description = ""
    with open("README.md", "r") as file:
        long_description += file.read()

    long_description += "\n\n"

    with open("CHANGELOG.md", "r") as file:
        long_description += file.read()

    return long_description


here = os.path.abspath(os.path.dirname(__file__))


def find_version(*file_paths):
    path = os.path.join(here, *file_paths)
    with codecs.open(path, 'r') as fp:
        version_file = fp.read()
        version_match = re.search(r"__version__ = ['\"]([^'\"]*)['\"]",
                                  version_file, re.M)
        if version_match:
            return version_match.group(1)
        raise RuntimeError("Unable to find version string.")


def find_auther(*file_paths):
    path = os.path.join(here, *file_paths)
    with codecs.open(path, 'r') as fp:
        auther_file = fp.read()
        auther_match = re.search(r"__auther__ = ['\"]([^'\"]*)['\"]",
                                 auther_file, re.M)
        if auther_match:
            return auther_match.group(1)
        raise RuntimeError("Unable to find auther string.")


setup(
    name="wappstoiot",
    python_requires='>=3.7.0',
    version=find_version("wappstoiot", "__init__.py"),
    author=find_auther("wappstoiot", "__init__.py"),
    author_email="support@seluxit.com",
    license="Apache-2.0",
    description="Simple Wappsto Python user-interface to Wappsto IoT",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/Wappsto/python-wappsto-iot",
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    # tests_require=[
    #     'pytest',
    #     'tox'
    # ],
    package_data={
        'wappstoiot': ["py.typed", "*.pyi", "**/*.pyi"]
    },
    install_requires=[
        'slxjsonrpc>=0.9.2',
        'pydantic>=2.0.0,<3.0.0',
    ],
    # entry_points={  # TODO: fix __main__.py to be optional.
    #     "console_scripts": "wappstoiot=wappstoiot:__main__"
    # },
    # extras_require={  # TODO: fix __main__.py to be optional.
    #     "cli": [
    #         'requests',
    #     ]
    # },
)
